#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for PrepareLibs."""
from __future__ import absolute_import
import unittest
import sys
from mock import patch, MagicMock as Mock

from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc, main
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertMockCalls

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.PrepareLibs'


class TestPrepareLibs(unittest.TestCase):
  """Tests the removeLibc method."""

  def setUp(self):
    sys_mock = Mock()
    sys_mock.argv = ['something', 'myothertestpath']
    mocked_modules = {'sys': sys_mock}
    self.module_patcher = patch.dict(sys.modules, mocked_modules)
    self.module_patcher.start()

  def tearDown(self):
    self.module_patcher.stop()

  def test_remove_libc(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'mytestpath', 'mytestpath', 'mytestpath'])) as getcwd_mock, \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])) as listdir_mock:
      result = removeLibc('mytestpath')
      self.assertTrue(result)
      assertMockCalls(chdir_mock, ['mytestpath', 'current_dir'], self)
      listdir_mock.assert_called_once_with('mytestpath')
      assertMockCalls(remove_mock, ['mytestpath/libc.so', 'mytestpath/libstdc++.so'], self)
      assertMockCalls(getcwd_mock, [()] * 4, self)

  def test_remove_libc_chdir_fails(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'mytestpath', 'mytestpath', 'mytestpath'])) as getcwd_mock, \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(side_effect=OSError('chdir_test_os_err'))) as chdir_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])) as listdir_mock:
      result = removeLibc('mytestpath')
      self.assertTrue(result)
      chdir_mock.assert_called_once_with('mytestpath')
      self.assertFalse(listdir_mock.called)
      self.assertFalse(remove_mock.called)
      getcwd_mock.assert_called_once_with()

  def test_remove_libc_remove_fails(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'mytestpath', 'mytestpath', 'mytestpath'])) as getcwd_mock, \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=OSError('test_cannot_remove_os_err'))) as remove_mock, \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])) as listdir_mock:
      result = removeLibc('mytestpath')
      self.assertFalse(result)
      assertMockCalls(chdir_mock, ['current_dir', 'mytestpath'], self)
      listdir_mock.assert_called_once_with('mytestpath')
      remove_mock.assert_called_once_with('mytestpath/libc.so')
      assertMockCalls(getcwd_mock, [()] * 3, self)

  def test_remove_libc_nothing_to_remove(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'mytestpath', 'mytestpath', 'mytestpath'])) as getcwd_mock, \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)) as chdir_mock, \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)) as remove_mock, \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'not_a_library.txt', 'dont_delete_me_libsomething.so', 'innocentlibc_.so'])) as listdir_mock:
      result = removeLibc('mytestpath')
      self.assertTrue(result)
      assertMockCalls(chdir_mock, ['mytestpath', 'current_dir'], self)
      listdir_mock.assert_called_once_with('mytestpath')
      self.assertFalse(remove_mock.called)
      assertMockCalls(getcwd_mock, [()] * 2, self)

  def test_main(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'myothertestpath', 'myothertestpath', 'myothertestpath'])), \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)), \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)), \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])):
      assertEqualsImproved(main(), 0, self)

  def test_main_no_args(self):
    self.module_patcher.stop()
    sys_mock = Mock()
    sys_mock.argv = ['something']
    mocked_modules = {'sys': sys_mock}
    self.module_patcher = patch.dict(sys.modules, mocked_modules)
    self.module_patcher.start()
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'myothertestpath', 'myothertestpath', 'myothertestpath'])), \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.remove' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])):
      assertEqualsImproved(main(), 1, self)

  def test_main_remove_fails(self):
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(side_effect=['current_dir', 'myothertestpath', 'myothertestpath', 'myothertestpath'])), \
        patch('%s.os.chdir' % MODULE_NAME, new=Mock(return_value=True)), \
        patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=OSError('test_cannot_remove_os_err'))), \
        patch('%s.os.listdir' % MODULE_NAME, new=Mock(return_value=['directory_content1.txt', 'libc.so', 'libstdc++.so'])):
      assertEqualsImproved(main(), 1, self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Unit tests for the CalibrationClient."""

from __future__ import absolute_import
import pytest
from mock import MagicMock as Mock
from DIRAC import S_OK
from ILCDIRAC.CalibrationSystem.Client.CalibrationClient import CalibrationClient

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.CalibrationSystem.Client.CalibrationClient'


@pytest.fixture
def calibClient():
  """Create calibration handler."""
  calibClient = CalibrationClient(1, 1)
  return calibClient


def test_getInputDataDict(mocker):
  """Test getInputDataDict."""
  #  mocker.patch('%s.Client' % MODULE_NAME, new=Mock(return_value=True))
  from DIRAC.Core.Base.Client import Client
  mocker.patch.object(Client, '__init__', new=Mock(return_value=True))

  calibClientWithInputArguments = CalibrationClient(1, 1)

  tmpMock = Mock(name='instance')
  tmpMock.getInputDataDict.return_value = S_OK()
  mocker.patch.object(calibClientWithInputArguments, '_getRPC', return_value=tmpMock)

  res = calibClientWithInputArguments .getInputDataDict()
  assert res['OK']

  calibClientWoInputArguments = CalibrationClient()
  mocker.patch.object(calibClientWoInputArguments, '_getRPC', return_value=tmpMock)

  res = calibClientWoInputArguments .getInputDataDict(1, 1)
  assert res['OK']



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Several file utilities.

:since: Mar 13, 2013
:author: sposs
"""

from __future__ import absolute_import
import glob
import os
import re
import shutil

from distutils import dir_util, errors  # pylint: disable=no-name-in-module

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.Operation import File
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


def upload(path, appTar):
  """Upload software tar ball to storage."""
  datMan = DataManager()
  ops = Operations()
  if path[-1] != "/":
    path += "/"
  if not os.path.exists(appTar):
    LOG.error("Tar ball %s does not exists, cannot continue." % appTar)
    return S_ERROR()
  if path.find("http://www.cern.ch/lcd-data") > -1:
    final_path = "/afs/cern.ch/eng/clic/data/software/"
    try:
      shutil.copy(appTar, "%s%s" % (final_path, os.path.basename(appTar)))
    except EnvironmentError as x:
      LOG.error("Could not copy because %s" % x)
      return S_ERROR("Could not copy because %s" % x)
  elif path.find("http://") > -1:
    LOG.error("Path %s was not foreseen!" % path)
    LOG.error("Location not known, upload to location yourself, and publish in CS manually")
    return S_ERROR()
  else:
    lfnpath = "%s%s" % (path, os.path.basename(appTar))
    res = datMan.putAndRegister(lfnpath, appTar, ops.getValue('Software/BaseStorageElement', "CERN-SRM"))
    if not res['OK']:
      return res
    request = Request()
    requestClient = ReqClient()
    request.RequestName = 'copy_%s' % os.path.basename(appTar).replace(".tgz", "").replace(".tar.gz", "")
    request.SourceComponent = 'ReplicateILCSoft'
    copies_at = ops.getValue('Software/CopiesAt', [])
    for copies in copies_at:
      transfer = Operation()
      transfer.Type = "ReplicateAndRegister"
      transfer.TargetSE = copies
      trFile = File()
      trFile.LFN = lfnpath
      trFile.GUID = ""
      transfer.addFile(trFile)
      request.addOperation(transfer)

    res = RequestValidator().validate(request)
    if not res['OK']:
      return res

    if copies_at:
      res = requestClient.putRequest(request)
      if not res['OK']:
        LOG.error('Could not set replication request', res['Message'])
      return S_OK('Application uploaded')
  return S_OK()


def fullCopy(srcdir, dstdir, item):
  """Copy the item from srcdir to dstdir, creates missing directories if needed."""
  item = item.rstrip().lstrip().lstrip("./").rstrip("/")
  srcdir = srcdir.rstrip("/")
  dstdir = dstdir.rstrip("/")
  if not re.match(r"(.*)[a-zA-Z0-9]+(.*)", item):  # we want to have explicit elements
    LOG.error("You try to get all files, that cannot happen")
    return S_OK()
  src = os.path.join(srcdir, item)
  items = glob.glob(src)
  if not items:
    LOG.error("No items found matching", src)
    return S_ERROR("No items found!")

  for item in items:
    item = item.replace(srcdir, "").lstrip("/")
    dst = os.path.join(dstdir, item)

    try:
      dir_util.create_tree(dstdir, [item])
    except errors.DistutilsFileError as why:
      return S_ERROR(str(why))

    if os.path.isfile(os.path.join(srcdir, item)):
      try:
        shutil.copy2(os.path.join(srcdir, item), dst)
      except EnvironmentError as why:
        return S_ERROR(str(why))
    else:
      try:
        shutil.copytree(os.path.join(srcdir, item), dst)
      except EnvironmentError as why:
        return S_ERROR(str(why))
  return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Run the StdHepCut utility.

Apply a set of cuts on input stdhep files

:since: May 11, 2011
:author: Stephane Poss
"""

from __future__ import absolute_import
import os
import shutil

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import shellCall

from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder
from ILCDIRAC.Core.Utilities.PrepareOptionFiles import getNewLDLibs
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDirName
from ILCDIRAC.Core.Utilities.resolvePathsAndNames import getProdFilename

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class StdHepCut(ModuleBase):
  """Apply cuts on stdhep files, based on L.

  Weuste utility.
  """

  def __init__(self):
    super(StdHepCut, self).__init__()
    self.applicationName = 'stdhepCut'
    self.STEP_NUMBER = ''
    self.inlineCuts = ""
    self.MaxNbEvts = 0
    self.scriptName = ""
    self.fileMask = '*.stdhep'

  def applicationSpecificInputs(self):

    if not self.OutputFile:
      dircont = os.listdir("./")
      for myfile in dircont:
        if myfile.count(".stdhep"):
          self.OutputFile = myfile.replace(".stdhep", "_reduced.stdhep")
          break
      if not self.OutputFile:
        return S_ERROR("Could not find suitable OutputFile name")

    if "IS_PROD" in self.workflow_commons and self.workflow_commons["IS_PROD"]:
      # self.OutputFile = getProdFilename(self.OutputFile,int(self.workflow_commons["PRODUCTION_ID"]),
      #                                  int(self.workflow_commons["JOB_ID"]))
      if 'ProductionOutputData' in self.workflow_commons:
        outputlist = self.workflow_commons['ProductionOutputData'].split(";")
        for obj in outputlist:
          if obj.lower().count("_gen_"):
            self.OutputFile = os.path.basename(obj)
            break
      else:
        self.OutputFile = getProdFilename(self.OutputFile,
                                          int(self.workflow_commons["PRODUCTION_ID"]),
                                          int(self.workflow_commons["JOB_ID"]))

    LOG.notice("Outputfile: %s" % self.OutputFile)
    if self.inlineCuts:
      cfile = open("cuts_local.txt", "w")
      cfile.write("\n".join(self.inlineCuts.split(";")))
      cfile.close()
      self.SteeringFile = "cuts_local.txt"
    return S_OK()

  def runIt(self):
    """Called from Workflow."""

    if not self.workflowStatus['OK'] or not self.stepStatus['OK']:
      LOG.verbose('Workflow status = %s, step status = %s' % (self.workflowStatus['OK'], self.stepStatus['OK']))
      return S_OK('%s should not proceed as previous step did not end properly' % self.applicationName)

    if not self.OutputFile:
      LOG.error("OutputFile not specified")
      return S_ERROR("OutputFile not specified")

    res = getSoftwareFolder(self.platform, self.applicationName, self.applicationVersion)
    if not res['OK']:
      LOG.error('Application %s was not found in either the local area or shared area' % self.applicationName)
      self.setApplicationStatus('%s: Could not find neither local area not shared area install' % self.applicationName)
      return res
    mySoftDir = res['Value']

    self.SteeringFile = os.path.basename(self.SteeringFile)
    if not os.path.exists(self.SteeringFile):
      LOG.verbose('Getting the steering files directory')
      res = getSteeringFileDirName(self.platform, self.applicationName, self.applicationVersion)
      if not res['OK']:
        LOG.error("Could not locate the steering file directory")
        return res
      steeringfiledirname = res['Value']
      if os.path.exists(os.path.join(steeringfiledirname, self.SteeringFile)):
        try:
          shutil.copy(os.path.join(steeringfiledirname, self.SteeringFile), "./" + self.SteeringFile)
        except EnvironmentError as x:
          LOG.error("Failed to get the cuts file")
          return S_ERROR('Failed to access file %s: %s' % (self.SteeringFile, str(x)))
    cuts = open(self.SteeringFile, "r")
    cutslines = "".join(cuts.readlines())
    cuts.close()
    LOG.verbose("Content of cuts file: ", cutslines)

    # Create the cut specific run script. Overloaded in the StdhepCutJava
    self.prepareScript(mySoftDir)

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)
    os.chmod(self.scriptName, 0o755)
    comm = 'sh -c "./%s"' % (self.scriptName)
    self.setApplicationStatus('%s %s step %s' % (self.applicationName, self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''
    self.result = shellCall(0, comm, callbackFunction=self.redirectLogOutput, bufferLimit=20971520)
    #self.result = {'OK':True,'Value':(0,'Disabled Execution','')}
    resultTuple = self.result['Value']
    if not os.path.exists(self.applicationLog):
      LOG.error("Something went terribly wrong, the log file is not present")
      self.setApplicationStatus('%s failed terribly, you are doomed!' % (self.applicationName))
      if not self.ignoreapperrors:
        LOG.error('Missing log file')
        return S_ERROR('%s did not produce the expected log' % (self.applicationName))
    status = resultTuple[0]
    # stdOutput = resultTuple[1]
    # stdError = resultTuple[2]
    LOG.info("Status after the application execution is %s" % str(status))

    nbevtswritten = -1
    nbevtspassing = 0
    nbevtsread = 0
    with open(self.applicationLog, 'r') as logf:
      for line in logf:
        line = line.rstrip()
        if line.count('Events kept'):
          nbevtswritten = int(line.split()[-1])
        if line.count('Events passing cuts'):
          nbevtspassing = int(line.split()[-1])
        if line.count('Events total'):
          nbevtsread = int(line.split()[-1])
    if nbevtswritten > 0 and nbevtspassing > 0 and nbevtsread > 0:
      cut_eff = 1. * nbevtspassing / nbevtsread
      LOG.info('Selection cut efficiency : %s%%' % (100 * cut_eff))
      sel_eff = 1. * nbevtswritten / nbevtspassing
      if nbevtswritten < self.MaxNbEvts:
        LOG.error('Not enough events to fill up')
      if 'Luminosity' in self.workflow_commons:
        self.workflow_commons['Luminosity'] = self.workflow_commons['Luminosity'] * sel_eff
      self.workflow_commons['NbOfEvts'] = nbevtswritten
      info = {}
      info['stdhepcut'] = {}
      info['stdhepcut']['Reduction'] = sel_eff
      info['stdhepcut']['CutEfficiency'] = cut_eff
      if 'Info' not in self.workflow_commons:
        self.workflow_commons['Info'] = info
      else:
        self.workflow_commons['Info'].update(info)
    else:
      LOG.error('Not enough events somewhere: read: %s, pass:%s, written:%s' % (nbevtsread, nbevtspassing,
                                                                                     nbevtswritten))
      status = 1

    return self.finalStatusReport(status)

  def prepareScript(self, mySoftDir):
    """Prepare the script."""
    new_ld_lib_path = getNewLDLibs(self.platform, self.applicationName, self.applicationVersion)
    new_ld_lib_path = mySoftDir + "/lib:" + new_ld_lib_path
    if os.path.exists("./lib"):
      new_ld_lib_path = "./lib:" + new_ld_lib_path

    self.scriptName = '%s_%s_Run_%s.sh' % (self.applicationName, self.applicationVersion, self.STEP_NUMBER)
    if os.path.exists(self.scriptName):
      os.remove(self.scriptName)
    script = open(self.scriptName, 'w')
    script.write('#!/bin/sh \n')
    script.write('#####################################################################\n')
    script.write('# Dynamically generated script to run a production or analysis job. #\n')
    script.write('#####################################################################\n')
    script.write('declare -x PATH=%s:$PATH\n' % mySoftDir)
    script.write('declare -x LD_LIBRARY_PATH=%s\n' % new_ld_lib_path)
    script.write('env | sort >> localEnv.log\n')
    script.write('echo =============================\n')
    extraopts = ""
    if self.MaxNbEvts:
      extraopts = '-m %s' % self.MaxNbEvts
    comm = 'stdhepCut %s -o %s -c %s %s\n' % (extraopts, self.OutputFile, self.SteeringFile, self.fileMask)
    LOG.info("Running %s" % comm)
    script.write(comm)
    script.write('declare -x appstatus=$?\n')
    script.write('exit $appstatus\n')
    script.close()



"""Retrieve ECalToHad and HCalToHad from Calibration.txt file."""

from __future__ import print_function
from __future__ import absolute_import
import sys
from Helper_Functions import find_between


Calibration_File_And_Path = sys.argv[1]
Energy_To_Calibrate = sys.argv[2]
ECTH_or_HCTH = sys.argv[3]
Initial_Calibration_Constant = sys.argv[4]
FOM_Or_Calibration_Constant = sys.argv[5]
CSM_or_TEM = sys.argv[6]

KaonL_Mass = 0.497672

Kinetic_Energy_To_Calibrate = float(Energy_To_Calibrate) - KaonL_Mass

# Intercept is the last value for intercept written to the Calibration.txt file.

CSM_Intercept = ''
TEM_Multiplier = ''

if CSM_or_TEM == 'CSM':
    if ECTH_or_HCTH == 'ECTH':
        with open(Calibration_File_And_Path, 'r') as f:
            searchlines = f.readlines()
            for line in searchlines:
                if 'm_eCalToHadInterceptMinChi2' in line:
                    CSM_Intercept = float(find_between(line, ' : ', ' : '))

    elif ECTH_or_HCTH == 'HCTH':
        with open(Calibration_File_And_Path, 'r') as f:
            searchlines = f.readlines()
            for line in searchlines:
                if 'm_hCalToHadInterceptMinChi2' in line:
                    CSM_Intercept = float(find_between(line, ' : ', ' : '))

    else:
        print('Please select ECTH or HCTH CSM_Intercept to extract.')

    if FOM_Or_Calibration_Constant == 'FOM':
        print(CSM_Intercept)

    elif FOM_Or_Calibration_Constant == 'Calibration_Constant':
        print(str(float(Kinetic_Energy_To_Calibrate) * float(Initial_Calibration_Constant) / float(CSM_Intercept)))

        Calibration_Text = '___________________________________________________________________________________' + '\n'
        Calibration_Text += 'Had_Extract.py extracting '
        if ECTH_or_HCTH == 'HCTH':
            Calibration_Text += 'HCalToHad'
        elif ECTH_or_HCTH == 'ECTH':
            Calibration_Text += 'ECalToHad'
        Calibration_Text += ' from CSM PandoraPFA calibration program. ' + '\n'
        Calibration_Text += 'Kinetic Energy To Calibrate                        : ' + \
            str(Kinetic_Energy_To_Calibrate) + '\n'
        Calibration_Text += 'Initial Calibration Constant                       : ' + \
            str(Initial_Calibration_Constant) + '\n'
        Calibration_Text += 'CSM_Intercept                                      : ' + str(CSM_Intercept) + '\n\n'
        Calibration_Text += 'New Calibration Constant                           : ' + \
            str(float(Kinetic_Energy_To_Calibrate) * float(Initial_Calibration_Constant) / float(CSM_Intercept)) \
            + '\n\n'

        with open(Calibration_File_And_Path, 'a') as myfile:
            myfile.write(Calibration_Text)

    else:
        print('Please select Intercept or Calibration_Constant to extract.')

elif CSM_or_TEM == 'TEM':

    if ECTH_or_HCTH == 'ECTH':
        with open(Calibration_File_And_Path, 'r') as f:
            searchlines = f.readlines()
            for line in searchlines:
                if 'Minimum_RMS_ECal_Multiplier' in line:
                    TEM_Multiplier = float(find_between(line, ' : ', ' :'))

    elif ECTH_or_HCTH == 'HCTH':
        with open(Calibration_File_And_Path, 'r') as f:
            searchlines = f.readlines()
            for line in searchlines:
                if 'Minimum_RMS_HCal_Multiplier' in line:
                    TEM_Multiplier = float(find_between(line, ' : ', ' :'))

    else:
        print('Please select ECTH or HCTH TEM_Multiplier to extract.')

    if FOM_Or_Calibration_Constant == 'FOM':
        print(TEM_Multiplier)

    elif FOM_Or_Calibration_Constant == 'Calibration_Constant':
        print(str(float(TEM_Multiplier) * float(Initial_Calibration_Constant)))

        Calibration_Text = '___________________________________________________________________________________' + '\n'
        Calibration_Text += 'Had_Extract.py extracting '
        if ECTH_or_HCTH == 'HCTH':
            Calibration_Text += 'HCalToHad'
        elif ECTH_or_HCTH == 'ECTH':
            Calibration_Text += 'ECalToHad'
        Calibration_Text += ' from TEM PandoraPFA calibration program. ' + '\n'
        Calibration_Text += 'Kinetic Energy To Calibrate                        : ' + \
            str(Kinetic_Energy_To_Calibrate) + '\n'
        Calibration_Text += 'Initial Calibration Constant                       : ' + \
            str(Initial_Calibration_Constant) + '\n'
        Calibration_Text += 'TEM_Multiplier                                     : ' + str(TEM_Multiplier) + '\n\n'
        Calibration_Text += 'New Calibration Constant                           : ' + \
            str(float(TEM_Multiplier) * float(Initial_Calibration_Constant)) + '\n\n'

        with open(Calibration_File_And_Path, 'a') as myfile:
            myfile.write(Calibration_Text)

    else:
        print('Please select FOM or Calibration_Constant to extract.')

else:
    print('Please select CSM or TEM.')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
DDSim: Simulation based on DD4hep and LCGeo

.. versionadded:: v25r0p0

Usage:

>>> ddsim = DDSim()
>>> ddsim.setVersion("ILCSoft-01-18-00")
>>> ddsim.setDetectorModel("CLIC_o2_v03")
>>> ddsim.setInputFile("LFN:/ilc/prod/clic/500gev/Z_uds/gen/0/00.stdhep")
>>> ddsim.setNumberOfEvents(30)

Use :func:`DDSim.setExtraCLIArguments` in case you want to use command line parameters


.. versionadded:: v26r0p7

To use your own plugins, detector drivers, or custom lcgeo and dd4hep libraries add a
tarball to the inputSandbox (via LFN) which contains the 'lib' folder with the
shared object and the components file::

   ./lib/liblcgeo.so
   ./lib/liblcgeo.components
   ./lib/...

"""
from __future__ import absolute_import
import types
import os

from ILCDIRAC.Interfaces.API.NewInterface.LCApplication import LCApplication
from ILCDIRAC.Core.Utilities.InstalledFiles import Exists
from ILCDIRAC.Interfaces.Utilities.DDInterfaceMixin import DDInterfaceMixin
from ILCDIRAC.Core.Utilities.Utilities import canConvert
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
import six
from six.moves import range

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class DDSim(DDInterfaceMixin, LCApplication):
  """DDSim Application Class."""

  def __init__(self, paramdict=None):
    self.startFrom = 0
    self.randomSeed = -1
    self.detectorModel = ''
    self.extraParticles = []
    super(DDSim, self).__init__(paramdict)
    # Those 5 need to come after default constructor
    self._modulename = 'DDSimAnalysis'
    self._moduledescription = 'Module to run DDSim'
    self.appname = 'ddsim'
    self.datatype = 'SIM'
    self._paramsToExclude.extend(["outputDstPath", "outputRecPath", "OutputDstFile", "OutputRecFile"])
    self._ops = Operations()

  @property
  def detectortype(self):
    """detectorType needed for transformations."""
    return self.detectorModel

  @detectortype.setter
  def detectortype(self, value):
    """ignore setting of detector type for ddsim."""
    pass

  def setRandomSeed(self, randomSeed):
    """ Optional: Define random seed to use. Default is the jobID.

    :param int randomSeed: Seed to use during simulation.
    """
    self._checkArgs({'randomSeed': int})
    self.randomSeed = randomSeed

  def setExtraParticles(self, extraParticles):
    """Optional: Add or modify the particles.tbl of DDG4.

    In case you want to modify default properties of particles or add particles that are not in the
    default DDSim particle.tbl.

    In case ``SIM.physics.pdgfile`` is specified in your steering file, this custom file will be modified,
    if the option is not set the file ``$DD4hep/DDG4/examples/particle.tbl`` from the chosen DD4hep version
    will be used.

    Example usage:

    >>> ddsim.setExtraParticles([[2000004,"susy-c_R", 2, 500, 1, 50],[2000005,"susy-b_R", 2, 1500, 1, 100]])

    or

    >>> ddsim.setExtraParticles("2000004 susy-c_R 2 500 1 50 2000005 susy-b_R 2 1500 1 100")

    :param extraParticles: If list entries should be e.g. [ID, "name", chg, mass, total width, lifetime]
                           If string entries should be whitespace separated e.g. "ID name chg mass total-width lifetime"
    :type extraParticles: list or str
    """
    if isinstance(extraParticles, str):
      extraParticles = extraParticles.split()
      if len(extraParticles) % 6 != 0:
        return self._reportError("The extraParticles string cannot be split up in multiples of 6")
      extraParticles = [extraParticles[x:x + 6] for x in range(0, len(extraParticles), 6)]
      for particle in extraParticles:
        condInt = canConvert(particle[0], int) and canConvert(particle[2], int)
        condFloat = canConvert(particle[3], float) and canConvert(particle[4], float) and canConvert(particle[5], float)
        if not (condInt and condFloat):
          return self._reportError("Cannot convert input string to proper format [int,str,int,float,float,float]")
        particle[0] = int(particle[0])
        particle[2] = int(particle[2])
        particle[3] = float(particle[3])
        particle[4] = float(particle[4])
        particle[5] = float(particle[5])

    self._checkArgs({'extraParticles': list})
    self.extraParticles = extraParticles

  def setStartFrom(self, startfrom):
    """ Optional: Define from where ddsim starts to read in the input file

    :param int startfrom: from where ddsim starts to read the input file
    """
    self._checkArgs({'startfrom': int})
    self.startFrom = startfrom

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

  def _prodjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _checkConsistency(self, job=None):
    """FIXME Check consistency of the DDSim application, this is called from the `Job` instance.

    :param job: The instance of the job
    :type job: ~ILCDIRAC.Interfaces.API.NewInterface.Job.Job
    :returns: S_OK/S_ERROR
    """

    parameterName = [pN for pN in job.workflow.parameters.getParametersNames() if 'ConfigPackage' in pN]
    if parameterName:
      LOG.notice("Found config parameter", parameterName)
      config = job.workflow.parameters.find(parameterName[0])
      configversion = config.value
    else:
      configversion = None
    # Platform must always be defined
    platform = job.workflow.parameters.find("Platform").value

    if not self.version:
      return S_ERROR('No version found')
    if self.steeringFile:
      if not os.path.exists(self.steeringFile) and not self.steeringFile.lower().startswith("lfn:"):
        res = Exists(self.steeringFile, platform=platform, configversion=configversion)
        if not res['OK']:
          return res

    if not self.detectorModel:
      return S_ERROR("No detectorModel set")

    for particle in self.extraParticles:
      if not isinstance(particle, (list, tuple)):
        return S_ERROR("Wrong format, extra particles not a list of lists")
      if len(particle) != 6:
        return S_ERROR("Particle property not of correct format "
                       "[ID, \"name\", chg, mass, total width, lifetime]")
      if not isinstance(particle[0], int):
        return S_ERROR("Particle property ID not int")
      if not isinstance(particle[1], str):
        return S_ERROR("Particle property name not string")
      if not isinstance(particle[2], int):
        return S_ERROR("Particle property chg not int")
      if not isinstance(particle[3], float):
        return S_ERROR("Particle property mass not float")
      if not isinstance(particle[4], float):
        return S_ERROR("Particle property total width not float")
      if not isinstance(particle[5], float):
        return S_ERROR("Particle property lifetime not float")

    #res = self._checkRequiredApp()
    # if not res['OK']:
    #  return res

    if self._jobtype != 'User':
      self._listofoutput.append({"outputFile": "@{OutputFile}", "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

      self.prodparameters['detectorType'] = self.detectortype
      # FIXME: Delete old code, detectorModel is checked for False already
      # if self.detectorModel:
      #self.prodparameters['slic_detectormodel'] = self.detectorModel
      self.prodparameters['slic_detectormodel'] = self.detectorModel

    if not self.startFrom:
      LOG.info('No startFrom defined for DDSim : start from the beginning')

    return S_OK()

  def _applicationModule(self):

    md1 = self._createModuleDefinition()
    md1.addParameter(Parameter("randomSeed", 0, "int", "", "", False, False,
                               "Random seed for the generator"))
    md1.addParameter(Parameter("detectorModel", "", "string", "", "", False, False,
                               "Detecor model for simulation"))
    md1.addParameter(Parameter("extraParticles", [], "list", "", "", False, False,
                               "Particles to be modified or added to particles.tbl"))
    md1.addParameter(Parameter("startFrom", 0, "int", "", "", False, False,
                               "From where DDSim starts to read the input file"))
    md1.addParameter(Parameter("debug", False, "bool", "", "", False, False, "debug mode"))
    return md1

  def _applicationModuleValues(self, moduleinstance):

    moduleinstance.setValue("randomSeed", self.randomSeed)
    moduleinstance.setValue("detectorModel", self.detectorModel)
    moduleinstance.setValue("extraParticles", self.extraParticles)
    moduleinstance.setValue("startFrom", self.startFrom)
    moduleinstance.setValue("debug", self.debug)

  def _checkWorkflowConsistency(self):
    return self._checkRequiredApp()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, six.integer_types):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""CalibrationRun class."""

from __future__ import absolute_import
import glob
import os
import threading
from collections import defaultdict
from datetime import datetime
import shutil

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
import ILCDIRAC.CalibrationSystem.Utilities as utilities
from ILCDIRAC.CalibrationSystem.Utilities.fileutils import binaryFileToString
from ILCDIRAC.CalibrationSystem.Utilities.functions import convert_and_execute
from ILCDIRAC.CalibrationSystem.Utilities.functions import searchFilesWithPattern
from ILCDIRAC.CalibrationSystem.Utilities.functions import readParameterDict
from ILCDIRAC.CalibrationSystem.Utilities.functions import readParametersFromSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import updateSteeringFile
from ILCDIRAC.CalibrationSystem.Utilities.functions import saveCalibrationRun
from ILCDIRAC.CalibrationSystem.Utilities.functions import addParameterToProcessor
from ILCDIRAC.CalibrationSystem.Utilities.mergePandoraLikelihoodData import mergeLikelihoods
from ILCDIRAC.Interfaces.API.NewInterface.UserJob import UserJob
from ILCDIRAC.Interfaces.API.DiracILC import DiracILC
from ILCDIRAC.Interfaces.API.NewInterface.Applications.Calibration import Calibration
from six.moves import range

# TODO do we need it here (since there is one in calibrationHandler file). What is this for?
__RCSID__ = "$Id$"
LOG = gLogger.getSubLogger(__name__)


class CalibrationPhase(object):
  """Represents the different phases a calibration can be in.

  Since Python 2 does not have enums, this is hardcoded for the moment.
  Should this solution not be sufficient any more, one can make a better enum implementation by hand or install
  a backport of the python3 implementation from PyPi.
  """

  ECalDigi, HCalDigi, MuonAndHCalOtherDigi, ElectroMagEnergy, HadronicEnergy, PhotonTraining = list(range(6))

  @staticmethod
  def phaseIDFromString(phase_name):
    """Return the ID of the given CalibrationPhase, passed as a string.

    :param str phase_name: Name of the CalibrationPhase. Allowed are:
                           ECalDigi, HCalDigi, MuonAndHCalOtherDigi,
                           ElectroMagEnergy, HadronicEnergy, PhotonTraining
    :returns: ID of this phase
    :rtype: int
    """
    if phase_name == 'ECalDigi':
      return 0
    elif phase_name == 'HCalDigi':
      return 1
    elif phase_name == 'MuonAndHCalOtherDigi':
      return 2
    elif phase_name == 'ElectroMagEnergy':
      return 3
    elif phase_name == 'HadronicEnergy':
      return 4
    elif phase_name == 'PhotonTraining':
      return 5
    else:
      raise ValueError('There is no CalibrationPhase with the name %s' % phase_name)

  @staticmethod
  def fileKeyFromPhase(phaseID):
    """Return the ID of the given CalibrationPhase, passed as a string.

    :param str phase_name: Name of the CalibrationPhase. Allowed are:
                           ECalDigi, HCalDigi, MuonAndHCalOtherDigi,
                           ElectroMagEnergy, HadronicEnergy, PhotonTraining
    :returns: file key for this phase
    :rtype: str
    """
    if phaseID == CalibrationPhase.ECalDigi:
      return "GAMMA"
    elif phaseID == CalibrationPhase.HCalDigi:
      return "KAON"
    elif phaseID == CalibrationPhase.MuonAndHCalOtherDigi:
      return "MUON"
    elif phaseID == CalibrationPhase.ElectroMagEnergy:
      return "GAMMA"
    elif phaseID == CalibrationPhase.HadronicEnergy:
      return "KAON"
    elif phaseID == CalibrationPhase.PhotonTraining:
      return "ZUDS"
    else:
      raise ValueError('There is no CalibrationPhase with the ID %s' % phaseID)

  # TODO read these energies from CS or from users input
  @staticmethod
  def sampleEnergyFromPhase(phaseID):
    """Return energy of provided sample of the given CalibrationPhase, passed as a float.

    :param str phase_name: Name of the CalibrationPhase. Allowed are:
                           ECalDigi, HCalDigi, MuonAndHCalOtherDigi,
                           ElectroMagEnergy, HadronicEnergy, PhotonTraining
    :returns: file key for this phase
    :rtype: str
    """
    if phaseID == CalibrationPhase.ECalDigi:
      return 10.0
    elif phaseID == CalibrationPhase.HCalDigi:
      return 50.0
    elif phaseID == CalibrationPhase.MuonAndHCalOtherDigi:
      return 10.0
    elif phaseID == CalibrationPhase.ElectroMagEnergy:
      return 10.0
    elif phaseID == CalibrationPhase.HadronicEnergy:
      return 50.0
    elif phaseID == CalibrationPhase.PhotonTraining:
      return 200.0
    else:
      raise ValueError('There is no CalibrationPhase with the ID %s' % phaseID)

  @staticmethod
  def phaseNameFromID(phaseID):
    """Return the name of the CalibrationPhase with the given ID, as a string.

    :param int phaseID: ID of the enquired CalibrationPhase
    :returns: The name of the CalibrationPhase
    :rtype: str
    """
    if phaseID == CalibrationPhase.ECalDigi:
      return 'ECalDigi'
    elif phaseID == CalibrationPhase.HCalDigi:
      return 'HCalDigi'
    elif phaseID == CalibrationPhase.MuonAndHCalOtherDigi:
      return 'MuonAndHCalOtherDigi'
    elif phaseID == CalibrationPhase.ElectroMagEnergy:
      return 'ElectroMagEnergy'
    elif phaseID == CalibrationPhase.HadronicEnergy:
      return 'HadronicEnergy'
    elif phaseID == CalibrationPhase.PhotonTraining:
      return 'PhotonTraining'
    else:
      raise ValueError('There is no CalibrationPhase with the name %d' % phaseID)


# pylint: disable=no-self-use
class CalibrationResult(object):
  """Wrapper class to store information about calibration computation interim results.

  Stores results from all worker nodes from a single step.
  """

  def __init__(self):
    """Initialize."""
    self.results = dict()

  def addResult(self, workerID, result):
    """Add a result from a given worker to the object.

    :param int workerID: ID of the worker providing the result
    :param result: list of floats representing the returned histogram
    :type result: `python:list`
    :returns: None
    """
    self.results[workerID] = result

  def getNumberOfResults(self):
    """Return number of interim results stored in this wrapper.

    :returns: Number of histograms stored in this object
    :rtype: int
    """
    return len(self.results)


class CalibrationRun(object):
  """Object that stores information about a single run of the calibration software.

  Includes files, current parameter set, software version, the workers running as well as
  the results of each step.
  """

  def __init__(self, calibrationID, inputFiles, calibSettingsDict):
    """Initialize."""
    self.lock = threading.Lock()
    self.calibrationID = calibrationID
    self.settings = calibSettingsDict
    self.inputFiles = inputFiles
    self.log = LOG.getSubLogger('[%s]' % calibrationID)
    if self.settings['steeringFile'].lower().startswith('lfn:'):
      self.settings['steeringFile'] = self.settings['steeringFile'].split(':')[1]
    self.localSteeringFile = os.path.join("calib%s/" % self.calibrationID,
                                          os.path.basename(self.settings['steeringFile']))
    self.stepResults = defaultdict(CalibrationResult)
    self.currentStage = self.settings['startStage']
    self.currentPhase = self.settings['startPhase']
    self.currentStep = 0
    self.currentParameterSet = defaultdict()
    # TODO temporary field in the settings. for testing only
    self.calibrationFinished = self.settings['startCalibrationFinished']
    #  self.calibrationFinished = False
    self.resultsSuccessfullyCopied = False
    self.timeStampPattern = "%Y-%m-%d %H:%M:%S"
    self.calibrationStartTime = datetime.now().strftime(self.timeStampPattern)
    self.calibrationEndTime = None
    self.newPhotonLikelihood = None
    self.ops = Operations()
    self.calibrationConstantsDict = None
    self.proxyUserName = ''
    self.proxyUserGroup = ''
    # DEBUG properties:
    self.stopStage = self.settings['stopStage']
    self.stopPhase = self.settings['stopPhase']
    self.nFailedJobs = 0
    self.calibrationRunStatus = 'Starting'

    # self.workerJobs = [] ##FIXME: Disabled because not used? Maybe in submit initial jobs
    # self.activeWorkers = dict() ## dict between calibration and worker node? ##FIXME:Disabled because not used?
    # FIXME: Probably need to store a mapping workerID -> part of calibration that worker is working on. This then needs
    # to be accessed by the agent in the case of resubmission

  def __getstate__(self):
    """Copy the object's state from self.__dict__ which contains all instance attributes.

    This implementatin is required to dump instance of calibrationRun to local file.
    """
    state = self.__dict__.copy()
    # Remove the unpicklable entries.
    state.pop('log', None)
    state.pop('ops', None)
    state.pop('lock', None)
    return state

  def __setstate__(self, state):
    """Restore the object's state from self.__dict__ which contains all instance attributes.

    This implementatin is required to dump instance of calibrationRun to local file.
    """
    self.__dict__.update(state)
    # Restore unpicklable entries
    self.ops = Operations()
    self.lock = threading.Lock()
    self.log = LOG.getSubLogger('[%s]' % self.calibrationID)

  def getCurrentStatus(self):
    """Return dict with calibration current status."""
    outDict = {}
    outDict['calibrationRunStatus'] = self.calibrationRunStatus
    outDict['calibrationID'] = self.calibrationID
    outDict['currentStage'] = self.currentStage
    outDict['currentPhase'] = self.currentPhase
    outDict['currentStep'] = self.currentStep
    outDict['calibrationStartTime'] = self.calibrationStartTime
    outDict['calibrationFinished'] = self.calibrationFinished
    outDict['resultsSuccessfullyCopied'] = self.resultsSuccessfullyCopied
    if self.calibrationEndTime is not None:
      outDict['calibrationEndTime'] = self.calibrationEndTime
    return outDict

  def checkForDebugFlagsToStopCalibration(self):
    """Check if calibration reached stopPhase and stopStage values."""
    if (self.stopStage < self.currentParameterSet['currentStage']):
      return True
    elif (self.stopStage == self.currentParameterSet['currentStage']
            and self.stopPhase < self.currentParameterSet['currentPhase']):
      return True
    else:
      return False

  def readInitialParameterDict(self):
    """Read initial parameters for calibration."""
    self.log.info('running readInitialParameterDict')

    dataMan = DataManager()
    res = dataMan.getFile(self.settings['steeringFile'], destinationDir='calib%s/' % self.calibrationID)

    if not res['OK'] or not os.path.exists(self.localSteeringFile):
      errMsgConst = 'Cannot copy Marlin steering file.'
      errMsgVariable = 'errMsg: %s' % res
      self.log.error(errMsgConst, errMsgVariable)
      return S_ERROR(errMsgConst + errMsgVariable)

    try:
      shutil.copyfile(self.localSteeringFile, "%s_INPUT" % self.localSteeringFile)
    except IOError:
      self.log.warn("Cannot make a copy of input steering file. This doesn't affect operation.",
                     "File name: %s_INPUT" % self.localSteeringFile)

    # TODO this path will be different in production version probably... update it
    parListFileName = os.path.join(utilities.__path__[0], 'auxiliaryFiles/parameterListMarlinSteeringFile.txt')

    parDict = readParameterDict(parListFileName)
    for key in list(parDict.keys()):
      newKey = None
      if 'MyDDMarlinPandora' in key:
        newKey = key.replace('MyDDMarlinPandora', self.settings['DDPandoraPFANewProcessorName'])
      if 'MyDDCaloDigi' in key:
        newKey = key.replace('MyDDCaloDigi', self.settings['DDCaloDigiName'])
      if newKey is not None:
        parDict[newKey] = parDict.pop(key)

    res = readParametersFromSteeringFile(self.localSteeringFile, parDict, ['PfoAnalysis'])
    if not res['OK']:
      self.log.error('Failed to read parameters from steering file:', res['Message'])
      return S_ERROR('Failed to read parameters from steering file')

    if self.settings['disableSoftwareCompensation']:
      tmpKey = (".//processor[@name='%s']/parameter[@name='MaxClusterEnergyToApplySoftComp']"
                % self.settings['DDPandoraPFANewProcessorName'])
      # add 'MaxClusterEnergyToApplySoftComp' parameter to calibrationConstantsDict (to be used on each worker node)
      parDict[tmpKey] = 0.0
      # setup 'MaxClusterEnergyToApplySoftComp' for local file
      tmpDict = {tmpKey: None}
      res = readParametersFromSteeringFile(self.localSteeringFile, tmpDict)
      # if no such parameter in the steering file --> add it
      if not res['OK']:
        res = addParameterToProcessor(self.localSteeringFile, self.settings['DDPandoraPFANewProcessorName'], {
                                      'name': 'MaxClusterEnergyToApplySoftComp', 'type': 'float', 'value': '0'})
        self.log.info("self.settings['DDPandoraPFANewProcessorName']:",
                      "%s" % self.settings['DDPandoraPFANewProcessorName'])
        self.log.info('out of addParameterToProcessor:', '%s' % res)
        self.log.info('self.localSteeringFile:', '%s' % self.localSteeringFile)
        if not res['OK']:
          self.log.error('Message from addParameterToProcessor function:', '%s' % res['Message'])

    if self.settings['nEcalThickLayers'] > 0:
      tmpKey = ".//processor[@name='%s']/parameter[@name='ECALLayers']" % self.settings['DDCaloDigiName']
      # +1 is just for safety
      valToSetup = '%s %s' % (self.settings['nEcalThinLayers'],
                              self.settings['nEcalThinLayers'] + self.settings['nEcalThickLayers'] + 1)
      # add 'ECALLayers' parameter to calibrationConstantsDict (to be used on each worker node)
      parDict[tmpKey] = valToSetup
      # setup 'ECALLayers' for local file
      tmpDict = {tmpKey: None}
      res = readParametersFromSteeringFile(self.localSteeringFile, tmpDict)
      # if no such parameter in the steering file --> add it
      if not res['OK']:
        res = addParameterToProcessor(self.localSteeringFile, self.settings['DDCaloDigiName'],
                                      {'name': 'ECALLayers', 'type': 'IntVec', 'value': valToSetup})
        if not res['OK']:
          self.log.error('Message from addParameterToProcessor function:', '%s' % res['Message'])
      else:
        res = updateSteeringFile(self.localSteeringFile, self.localSteeringFile,
                                 {tmpKey: valToSetup})
        if not res['OK']:
          self.log.error('Error while updating local steering file. Error message:', '%s' % res['Message'])
          return res

    self.calibrationConstantsDict = parDict
    self.currentParameterSet['currentStage'] = self.currentStage
    self.currentParameterSet['currentPhase'] = self.currentPhase
    self.currentParameterSet['currentStep'] = self.currentStep
    self.currentParameterSet['parameters'] = self.calibrationConstantsDict
    self.currentParameterSet['calibrationIsFinished'] = self.calibrationFinished

    return S_OK()

  def getCalibrationID(self):
    """Get calibration ID."""
    return self.calibrationID

  @executeWithUserProxy
  def submitJobs(self, idsOfWorkerNodesToSubmitTo=None):
    """Submit the calibration jobs to the workers for the first time.

    :param idsOfWorkerNodesToSubmitTo: list of integers representing IDs of worker nodes to submit jobs to;
                                       if None submit to all allocated nodes
    :returns: S_OK or S_ERROR
    :rtype: dict
    """
    self.log.info('running submitJobs')
    res = self.readInitialParameterDict()
    self.log.info('read initial parameter dict')
    if not res['OK']:
      #  self.log.error('Cannot read initial parameter dict. Message:', '%s' % res['Message'])
      errMsg = 'Cannot read initial parameter dict. Message:', '%s' % res['Message']
      self.log.info('Cannot read initial parameter dict. Probably wrong user input.')
      return S_ERROR(errMsg)

    dirac = DiracILC(True, 'calib%s/job_repository.rep' % self.calibrationID)
    results = []

    listOfNodesToSubmitTo = range(0, self.settings['numberOfJobs'])
    if idsOfWorkerNodesToSubmitTo is not None:
      listOfNodesToSubmitTo = idsOfWorkerNodesToSubmitTo

    key = CalibrationPhase.fileKeyFromPhase(self.currentPhase)
    self.log.verbose('fileKeyFromPhase:', '%s' % key)

    dirName = "calib%s/stage%s/phase%s/step%s" % (self.calibrationID,
                                                  self.currentStage, self.currentPhase, self.currentStep)
    if not os.path.exists(dirName):
      try:
        os.makedirs(dirName)
      except OSError as e:
        errMsgConst = 'Cannot create directory on service machine. Current working directory:'
        errMsgVariable = ': %s. Error message:%s' % (os.getcwd(), e)
        self.log.error(errMsgConst, errMsgVariable)
        return S_ERROR('System error. Contact administrator. Cannot create directory on the service machine.')

    for curWorkerID in listOfNodesToSubmitTo:
      # get input files
      fileList = []
      for _, iList in self.inputFiles[curWorkerID].items():
        fileList += iList[0]
      lcioFiles = fileList

      # create user job
      curJob = UserJob()
      curJob.check = False  # Necessary to turn off user confirmation
      curJob.setName('PandoraCaloCalibration_calid_%s_workerid_%s' % (self.calibrationID, curWorkerID))
      curJob.setJobGroup('PandoraCaloCalibration_calid_%s' % self.calibrationID)
      curJob.setConfigPackage(self.settings['nameOfTheConfigPackage'], self.settings['marlinVersion'].rsplit("_", 1)[0])

      # TODO implement using line below - choose of tracking, time window, etc.
      #  calib.setExtraCLIArguments(" --Config.Overlay="+overlayParameterValue+"  --Config.Tracking="+trackingType+"
      #                             --Output_DST.LCIOOutputFile="+outputFile+"
      #                             --constant.CalorimeterIntegrationTimeWindow="+str(calorimeterIntegrationTimeWindow))
      # FIXME use default CPU time limit?
      curJob.setCPUTime(24 * 60 * 60)
      # FIXME allow user to specify xml-files. CLIC detector have different name of PandoraLikelihhod file than CLD
      #  inputSB = ['GearOutput.xml', 'PandoraSettingsDefault.xml', 'PandoraLikelihoodData9EBin.xml']
      curJob.setInputSandbox(['LFN:' + self.settings['steeringFile']])
      curJob.setInputData(lcioFiles)
      # TODO files to redirect for output: newPhotonLikelihood.xml, finalSteeringFile
      curJob.setOutputSandbox(['*.log', '*.xml', '*.txt'])

      # create calibration workflow
      calib = Calibration()
      calib.setCalibrationID(self.calibrationID)
      calib.setWorkerID(curWorkerID)
      calib.setVersion(self.settings['marlinVersion'])
      calib.setDetectorModel(self.settings['detectorModel'])
      #  calib.setNbEvts(nEvts+1)
      #  calib.setProcessorsToUse([])
      calib.setSteeringFile(os.path.basename(self.settings['steeringFile']))
      res = curJob.append(calib)
      if not res['OK']:
        return S_ERROR('Failed to setup Calibration worklow module. WorkerID = %s; Error_msg: %s'
                       % (curWorkerID, res['Message']))

      # submit jobs
      res = curJob.submit(dirac, mode='wms')
      results.append(res)

    self.calibrationRunStatus = 'Running'
    saveCalibrationRun(self)

    return results

  def addResult(self, stepID, workerID, result):
    """Add a reconstruction result to the list of other results.

    :param int stepID: ID of the step
    :param int workerID: ID of the worker providing the result
    :param result: reconstruction histogram from the worker node
    :type result: `python:list`
    :returns: None
    """
    self.lock.acquire()
    self.stepResults[stepID].addResult(workerID, result)
    saveCalibrationRun(self)
    self.lock.release()
    # FIXME: Do we add old step results? Current status is no, ensured in CalibrationHandler
    # FIXME: Do we delete old interim results?

  def getNewParameters(self, stepIDOnWorker):
    """Return current parameters.

    :param int stepIDOnWorker: The ID of the step the worker just completed.
    :returns: If the computation is finished, returns S_OK containing a success message string. If there is a new
              parameter set, a S_OK dict containing the updated parameter set. Else a S_ERROR
    :rtype: dict
    """
    if self.currentStep > stepIDOnWorker:
      return S_OK(dict(self.currentParameterSet))
    else:
      self.log.verbose('No new parameter set available yet.', 'Current step in service: %s, step on worker: %s'
                       % (self.currentStep, stepIDOnWorker))
      return S_OK()

  def getNewPhotonLikelihood(self):
    """Return new photon likelihood."""
    if self.newPhotonLikelihood:
      return S_OK(self.newPhotonLikelihood)
    else:
      errMsgConst = 'No new parameter set available yet. Current step in service:'
      errMsgVariable = ('%s' % (self.currentStep))
      self.log.verbose(errMsgConst, errMsgVariable)
      return S_ERROR(errMsgConst + errMsgVariable)

  #  def __addLists(self, list1, list2):
  #    """Add two lists together by adding the first element, second element, and so on. Throws an exception
  #    if the lists have a different number of elements.
  #
  #    :param list1: List that should be added element-wise to another
  #    :type list1: `python:list`
  #    :param list2: Other list that should be added element-wise
  #    :type list2: `python:list`
  #    :returns: The list [list1[0]+list2[0], list1[1]+list2[1], ...]
  #    :rtype: list
  #    """
  #    if len(list1) != len(list2):
  #      raise ValueError('The two lists do not have the same number of elements. \n List 1: %s \n List 2: %s'
  #                       % (list1, list2))
  #    result = []
  #    for first_elem, second_elem in zip(list1, list2):
  #      result.append(first_elem + second_elem)
  #    return result

  def __mergePandoraLikelihoodXmlFiles(self):
    folder = "calib%s/stage%s/phase%s/" % (self.calibrationID, self.currentStage, self.currentPhase)
    if not os.path.exists(folder):
      return S_ERROR('no directory found: %s' % folder)

    filesToMerge = searchFilesWithPattern(folder, '*.xml')
    outFileName = "calib%s/newPandoraLikelihoodData.xml" % (self.calibrationID)

    res = mergeLikelihoods(filesToMerge, outFileName)
    if not res['OK']:
      return res

    if os.path.exists(outFileName):
      self.newPhotonLikelihood = binaryFileToString(outFileName)
      return S_OK()
    else:
      return S_ERROR('Failed to merge photon likelihoods')

  def getKey(self, pattern):
    """Return key from calibrationConstantsDict which contain input pattern."""
    for iKey in self.calibrationConstantsDict:
      if pattern in iKey:
        return iKey
    self.log.error('Cannot find XPath inside the parameter dict which contains pattern:', '%s' % pattern)
    return None

  def endCurrentStep(self):
    """Calculate new parameter set and prepare the object for the next step.

    :returns: None
    """
    self.log.info('Start execution of endCurrentStep.', 'Stage %s; Phase: %s; Step: %s' %
                  (self.currentStage, self.currentPhase, self.currentStep))

    if self.calibrationFinished:
      return S_ERROR('Calibration is finished. Do not call endCurrentStep() anymore!')

    fileNamePattern = 'pfoanalysis_w*.root'
    fileDir = "calib%s/stage%s/phase%s/step%s/" % (self.calibrationID, self.currentStage, self.currentPhase,
                                                   self.currentStep)
    inputFilesPattern = os.path.join(fileDir, fileNamePattern)
    inputFilesPattern = '"%s"' % inputFilesPattern  # needed for proper handling of wildcards
    fileDir = "calib%s/" % (self.calibrationID)
    calibrationFile = os.path.join(fileDir, "Calibration.txt")  # as hardcoded in calibration binaries

    self.log.info('calibrationFile:', '%s' % calibrationFile)

    scriptPath = self.ops.getValue("/AvailableTarBalls/%s/pandora_calibration_scripts/%s/%s"
                                   % (self.settings['platform'], self.settings['marlinVersion'], "PandoraAnalysis"),
                                   None)
    ilcSoftInitScript = self.ops.getValue("/AvailableTarBalls/%s/pandora_calibration_scripts/%s/%s"
                                          % (self.settings['platform'], self.settings['marlinVersion'],
                                             "CVMFSEnvScript"), None)

    pythonReadScriptPath = os.path.join(utilities.__path__[0], 'Python_Read_Scripts')

    truthEnergy = CalibrationPhase.sampleEnergyFromPhase(self.currentPhase)

    #  print('self.currentPhase', self.currentPhase)
    #  print('CalibrationPhase.ECalDigi', CalibrationPhase.ECalDigi)
    #  print('convert_and_execute', convert_and_execute())

    if self.currentPhase == CalibrationPhase.ECalDigi:
      binary = os.path.join(scriptPath, 'ECalDigitisation_ContainedEvents')

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['digitisationAccuracy'], '-d', fileDir, '-e', '90', '-g', 'Barrel',
                                 '-i', self.settings['ecalBarrelCosThetaRange'][0], '-j',
                                 self.settings['ecalBarrelCosThetaRange'][1]], ilcSoftInitScript)

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['digitisationAccuracy'], '-d', fileDir, '-e', '90', '-g', 'EndCap',
                                 '-i', self.settings['ecalEndcapCosThetaRange'][0], '-j',
                                 self.settings['ecalEndcapCosThetaRange'][1]], ilcSoftInitScript)

      # this parameter can be written in format "value value" in the xml steering file in case of ECAL layout which has
      # layers with different thicknesses
      prevStepCalibConstBarrel = float(self.calibrationConstantsDict[self.getKey('CalibrECAL')].split()[0])
      prevStepCalibConstEndcap = prevStepCalibConstBarrel * float(self.calibrationConstantsDict[
          self.getKey('ECALEndcapCorrectionFactor')])

      pythonReadScript = os.path.join(pythonReadScriptPath, 'ECal_Digi_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstBarrel, 'Calibration_Constant',
                                 'Barrel'])
      calibConstBarrel = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstEndcap, 'Calibration_Constant',
                                 'Endcap'])
      calibConstEndcap = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstBarrel, 'Mean', 'Barrel'])
      meanBarrel = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstEndcap, 'Mean', 'Endcap'])
      meanEndcap = float(res['Value'][1].split('\n')[0])

      self.calibrationConstantsDict[self.getKey('CalibrECAL')] = (
          '%s %s' % (calibConstBarrel, self.settings['ecalResponseCorrectionForThickLayers'] * calibConstBarrel))
      self.calibrationConstantsDict[self.getKey('ECALEndcapCorrectionFactor')] = (
          calibConstEndcap / calibConstBarrel)

      fractionalError = max(abs(meanBarrel - truthEnergy), abs(meanEndcap - truthEnergy)) / truthEnergy
      if fractionalError < self.settings['digitisationAccuracy']:
        self.currentPhase = self.currentPhase + 1

    elif self.currentPhase == CalibrationPhase.HCalDigi:
      binary = os.path.join(scriptPath, 'HCalDigitisation_ContainedEvents')
      kineticEnergy = truthEnergy - 0.497672

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['digitisationAccuracy'], '-d', fileDir, '-e', '90', '-g', 'Barrel',
                                 '-i', self.settings['hcalBarrelCosThetaRange'][0], '-j',
                                 self.settings['hcalBarrelCosThetaRange'][1]], ilcSoftInitScript)

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['digitisationAccuracy'], '-d', fileDir, '-e', '90', '-g', 'EndCap',
                                 '-i', self.settings['hcalEndcapCosThetaRange'][0], '-j',
                                 self.settings['hcalEndcapCosThetaRange'][1]], ilcSoftInitScript)

      prevStepCalibConstBarrel = float(self.calibrationConstantsDict[
          self.getKey('CalibrHCALBarrel')])
      prevStepCalibConstEndcap = float(self.calibrationConstantsDict[
          self.getKey('CalibrHCALEndcap')])

      pythonReadScript = os.path.join(pythonReadScriptPath, 'HCal_Digi_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kineticEnergy, prevStepCalibConstBarrel, 'Barrel',
                                 'Calibration_Constant'])
      calibConstBarrel = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kineticEnergy, prevStepCalibConstEndcap, 'EndCap',
                                 'Calibration_Constant'])
      calibConstEndcap = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kineticEnergy, prevStepCalibConstBarrel, 'Barrel', 'Mean'])
      meanBarrel = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kineticEnergy, prevStepCalibConstEndcap, 'EndCap', 'Mean'])
      meanEndcap = float(res['Value'][1].split('\n')[0])

      self.calibrationConstantsDict[
          self.getKey('CalibrHCALBarrel')] = calibConstBarrel
      self.calibrationConstantsDict[
          self.getKey('CalibrHCALEndcap')] = calibConstEndcap

      fractionalError = max(abs(meanBarrel - kineticEnergy), abs(meanEndcap - kineticEnergy)) / truthEnergy
      if fractionalError < self.settings['digitisationAccuracy']:
        self.currentPhase = self.currentPhase + 1

    elif self.currentPhase == CalibrationPhase.MuonAndHCalOtherDigi:
      binary = os.path.join(scriptPath, 'PandoraPFACalibrate_MipResponse')
      kineticEnergy = truthEnergy - 0.497672
      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', fileDir],
                                ilcSoftInitScript)

      pythonReadScript = os.path.join(pythonReadScriptPath, 'Extract_GeVToMIP.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'ECal'])
      ecalGevToMip = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'HCal'])
      hcalGevToMip = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'Muon'])
      muonGevToMip = float(res['Value'][1].split('\n')[0])
      # constants below were not used anywhere but they were calculated in the previous calibration procedure
      #  ecalMipMpv = float(convert_and_execute(['python', 'Extract_SimCaloHitMIPMPV.py', calibrationFile,
      #                                          'ECal']))
      #  hcalMipMpv = float(convert_and_execute(['python', 'Extract_SimCaloHitMIPMPV.py', calibrationFile,
      #                                          'HCal']))

      self.calibrationConstantsDict[self.getKey('ECalToMipCalibration')] = (
          ecalGevToMip)
      self.calibrationConstantsDict[self.getKey('HCalToMipCalibration')] = (
          hcalGevToMip)
      self.calibrationConstantsDict[self.getKey('MuonToMipCalibration')] = (
          muonGevToMip)

      binary = os.path.join(scriptPath, 'SimCaloHitEnergyDistribution')
      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', fileDir],
                                ilcSoftInitScript)

      pythonReadScript = os.path.join(pythonReadScriptPath, 'HCal_Ring_Digi_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy])
      mipPeakRatio = float(res['Value'][1].split('\n')[0])

      # this binary need to access kaon files --> refer to previous stage and step
      kaonTruthEnergy = CalibrationPhase.sampleEnergyFromPhase(self.currentPhase - 1)
      kaonInputFilesPattern = os.path.join("calib%s/stage%s/phase%s/step%s/"
                                           % (self.calibrationID, self.currentStage, self.currentPhase - 1,
                                              self.currentStep - 1), fileNamePattern)
      binary = os.path.join(scriptPath, 'HCalDigitisation_DirectionCorrectionDistribution')
      res = convert_and_execute([binary, '-a', kaonInputFilesPattern, '-b', kaonTruthEnergy,
                                 '-c', fileDir],
                                ilcSoftInitScript)

      pythonReadScript = os.path.join(pythonReadScriptPath, 'HCal_Direction_Corrections_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kaonTruthEnergy])
      directionCorrectionRatio = float(res['Value'][1].split('\n')[0])
      calibHcalEndcap = float(self.calibrationConstantsDict[
          self.getKey('CalibrHCALEndcap')])
      # one need to access hcalMeanEndcap again (as in previous phase. Read it again from calibration file...
      # but it will also write output from script execution to the file again
      pythonReadScript = os.path.join(pythonReadScriptPath, 'HCal_Digi_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 kaonTruthEnergy, calibHcalEndcap, 'Endcap', 'Mean'])
      hcalMeanEndcap = float(res['Value'][1].split('\n')[0])

      # TODO deal with these hardcoded values
      Absorber_Thickness_EndCap = 20.0
      Scintillator_Thickness_Ring = 3.0
      Absorber_Thickness_Ring = 20.0
      Scintillator_Thickness_EndCap = 3.0

      Absorber_Scintillator_Ratio = ((Absorber_Thickness_EndCap * Scintillator_Thickness_Ring)
                                     / (Absorber_Thickness_Ring * Scintillator_Thickness_EndCap))

      kaonKineticEnergy = CalibrationPhase.sampleEnergyFromPhase(CalibrationPhase.HCalDigi) - 0.49767
      calibHcalOther = (directionCorrectionRatio * mipPeakRatio * Absorber_Scintillator_Ratio * calibHcalEndcap
                        * kaonKineticEnergy / hcalMeanEndcap)

      self.calibrationConstantsDict[
          self.getKey('CalibrHCALOther')] = calibHcalOther

      self.currentPhase = self.currentPhase + 1

    elif self.currentPhase == CalibrationPhase.ElectroMagEnergy:
      binary = os.path.join(scriptPath, 'PandoraPFACalibrate_EMScale')

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['pandoraPFAAccuracy'], '-d', fileDir, '-e', '90'],
                                ilcSoftInitScript)

      prevStepCalibConstEcalToEm = float(self.calibrationConstantsDict[
          self.getKey('ECalToEMGeVCalibration')])
      pythonReadScript = os.path.join(pythonReadScriptPath, 'EM_Extract.py')
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstEcalToEm, 'Calibration_Constant'])
      ecalToEm = float(res['Value'][1].split('\n')[0])
      hcalToEm = ecalToEm
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, prevStepCalibConstEcalToEm, 'Mean'])
      emMean = float(res['Value'][1].split('\n')[0])

      self.calibrationConstantsDict[
          self.getKey('ECalToEMGeVCalibration')] = ecalToEm
      self.calibrationConstantsDict[
          self.getKey('HCalToEMGeVCalibration')] = hcalToEm

      fractionalError = abs(truthEnergy - emMean) / truthEnergy
      if fractionalError < self.settings['pandoraPFAAccuracy']:
        self.currentPhase += 1

    elif self.currentPhase == CalibrationPhase.HadronicEnergy:
      binary = os.path.join(scriptPath, 'PandoraPFACalibrate_HadronicScale_ChiSquareMethod')

      res = convert_and_execute([binary, '-a', inputFilesPattern, '-b', truthEnergy,
                                 '-c', self.settings['pandoraPFAAccuracy'], '-d', fileDir, '-e',
                                 self.settings['nHcalLayers']], ilcSoftInitScript)

      pythonReadScript = os.path.join(pythonReadScriptPath, 'Had_Extract.py')
      prevStepCalibConstHcalToHad = float(self.calibrationConstantsDict[
          self.getKey('HCalToHadGeVCalibration')])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'HCTH', prevStepCalibConstHcalToHad,
                                 'Calibration_Constant', 'CSM'])
      hcalToHad = float(res['Value'][1].split('\n')[0])
      prevStepCalibConstEcalToHad = float(self.calibrationConstantsDict[
          self.getKey('ECalToHadGeVCalibrationBarrel')])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'ECTH', prevStepCalibConstEcalToHad,
                                 'Calibration_Constant', 'CSM'])
      ecalToHad = float(res['Value'][1].split('\n')[0])

      self.calibrationConstantsDict[
          self.getKey('HCalToHadGeVCalibration')] = hcalToHad
      self.calibrationConstantsDict[
          self.getKey('ECalToHadGeVCalibrationBarrel')] = ecalToHad
      self.calibrationConstantsDict[
          self.getKey('ECalToHadGeVCalibrationEndCap')] = ecalToHad

      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'HCTH', hcalToHad, 'FOM', 'CSM'])
      hcalToHadFom = float(res['Value'][1].split('\n')[0])
      res = convert_and_execute(['python', pythonReadScript, calibrationFile,
                                 truthEnergy, 'ECTH', ecalToHad, 'FOM', 'CSM'])
      ecalToHadFom = float(res['Value'][1].split('\n')[0])

      kineticEnergy = truthEnergy - 0.497672
      fractionalError = max(abs(hcalToHadFom - kineticEnergy), abs(ecalToHadFom - kineticEnergy)) / kineticEnergy

      if fractionalError < self.settings['pandoraPFAAccuracy']:
        if self.currentStage == 1:
          self.currentStage += 1
          self.currentPhase += 1
        elif self.currentStage == 3:
          self.calibrationFinished = True
          self.calibrationRunStatus = "Calibration is succesfully finished."
          self.calibrationEndTime = datetime.now().strftime(self.timeStampPattern)
          self.log.info('The last step of calibration has been finished')
        else:
          return S_ERROR('%s' % self.currentStage)
    elif self.currentPhase == CalibrationPhase.PhotonTraining and self.currentStage == 2:
      res = self.__mergePandoraLikelihoodXmlFiles()
      if not res['OK']:
        return res
      self.currentStage += 1
      self.currentPhase = CalibrationPhase.ECalDigi
    else:
      return S_ERROR('Error in the execution sequence. Dump current ids\nstageID: %s, phaseID: %s, stepID: %s'
                     % (self.currentStage, self.currentPhase, self.currentStep))

    self.currentStep += 1

    self.currentParameterSet['currentStage'] = self.currentStage
    self.currentParameterSet['currentPhase'] = self.currentPhase
    self.currentParameterSet['currentStep'] = self.currentStep
    self.currentParameterSet['parameters'] = self.calibrationConstantsDict
    self.currentParameterSet['calibrationIsFinished'] = self.calibrationFinished

    if not self.calibrationFinished and self.checkForDebugFlagsToStopCalibration():
      self.currentParameterSet['calibrationIsFinished'] = True
      self.calibrationFinished = True
      self.calibrationEndTime = datetime.now().strftime(self.timeStampPattern)
      self.calibrationRunStatus = ("Calibration is succesfully finished. It reached requested stopStage:"
                                   " %s and stopPhase: %s" % (self.stopStage, self.stopPhase))

    # update local steering file after every step. This file will be used if calibration service will be restarted and
    # some calibrations are still are not finished
    res = updateSteeringFile(self.localSteeringFile, self.localSteeringFile,
                             self.calibrationConstantsDict, ["PfoAnalysis"])
    saveCalibrationRun(self)

    return S_OK()

  @executeWithUserProxy
  def copyResults(self):
    """Copy output files from the calibration."""
    res = updateSteeringFile(self.localSteeringFile, self.localSteeringFile,
                             self.calibrationConstantsDict, ["PfoAnalysis"])
    if not res['OK']:
      self.log.error('Error while updating local steering file. Error message:', '%s' % res['Message'])
      return res

    filesToCopy = []
    filesToCopy.append(self.localSteeringFile)
    if os.path.exists("calib%s/newPandoraLikelihoodData.xml" % (self.calibrationID)):
      filesToCopy.append("calib%s/newPandoraLikelihoodData.xml" % (self.calibrationID))
    filesToCopy.append("calib%s/Calibration.txt" % (self.calibrationID))
    filesToCopy += glob.glob("calib%s/*.C" % (self.calibrationID))
    filesToCopy += glob.glob("calib%s/*.png" % (self.calibrationID))
    filesToCopy += glob.glob("calib%s/*_INPUT" % (self.calibrationID))

    self.log.info('Start copying output of the calibration to user directory:',
                  '%s on storage element: %s' % (self.settings['outputPath'], self.settings['outputSE']))
    self.log.info('Files to copy:', '%s' % filesToCopy)

    dm = DataManager()
    for iFile in filesToCopy:
      if not os.path.exists(iFile):
        errMsgConst = "File doesn't exist locally."
        errMsgVariable = "File name: %s" % iFile
        self.log.error(errMsgConst, errMsgVariable)
        return(S_ERROR(errMsgConst + errMsgVariable))
      if not os.path.isfile(iFile):
        errMsgConst = "File is not found."
        errMsgVariable = "%s is not a file" % iFile
        self.log.error(errMsgConst, errMsgVariable)
        return(S_ERROR(errMsgConst + errMsgVariable))

      lfn = os.path.join(self.settings['outputPath'], "calib%s" % (self.calibrationID), os.path.basename(iFile))
      localFile = iFile
      res = dm.putAndRegister(lfn, localFile, self.settings['outputSE'], None, overwrite=True)
      if not res['OK']:
        errMsg = 'Error while uploading results. Error message: %s' % res['Message']
        self.log.error(errMsg)
        return res

    self.log.info('Copying is finished')
    return S_OK()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the ModuleBase module."""

from __future__ import print_function
from __future__ import absolute_import
try:
  from StringIO import StringIO  # for Python 2
except ImportError:
  from io import StringIO  # for Python 3
import sys
import unittest
from mock import patch, mock_open, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase, generateRandomString
from Tests.Utilities.GeneralUtils import assertEqualsImproved, \
    assertDiracFailsWith, assertDiracSucceeds, assertDiracSucceedsWith, \
    assertDiracSucceedsWith_equals, assertMockCalls
import six

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.ModuleBase'


class ModuleBaseTestCase(unittest.TestCase):  # pylint: disable=too-many-public-methods
  """Test the ModuleBase module."""

  def setUp(self):
    # Mock out modules that spawn other threads
    mocked_modules = {'DIRAC.DataManagementSystem.Client.DataManager': Mock()}

    self.log_mock = Mock()

    self.patches = [
        patch.dict(sys.modules, mocked_modules),
        patch('%s.LOG' % MODULE_NAME, new=self.log_mock),
        ]

    for patcher in self.patches:
      patcher.start()

    self.moba = ModuleBase()
    # clear logging from constructor
    self.log_mock.reset_mock()

  def tearDown(self):
    for patcher in self.patches:
      patcher.stop()

  def test_randomstring(self):
    random_string_1 = generateRandomString()
    random_string_2 = generateRandomString()
    assertEqualsImproved(len(random_string_1), 8, self)
    assertEqualsImproved(len(random_string_2), 8, self)
    assert isinstance(random_string_1, six.string_types)
    assert isinstance(random_string_2, six.string_types)

  def test_constructor_fail(self):
    """Test constructor when failing."""
    with patch('%s.getProxyInfoAsString' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_error_testme'))):
      ModuleBase()
    self.log_mock.error.assert_called_once_with(
        'Could not obtain proxy information in module environment with message:\n', 'some_error_testme')

  def test_constructor_sucess(self):
    """Test constructor when succeeding."""
    with patch('%s.getProxyInfoAsString' % MODULE_NAME, new=Mock(return_value=S_OK('some_proxy_infos'))):
      ModuleBase()
    self.assertFalse(self.log_mock.error.called)

  def test_execute_basic(self):
    result = self.moba.execute()
    assertDiracSucceeds(result, self)

  def test_execute_resolveinput_fails(self):
    self.moba.isProdJob = True
    self.moba.InputData = 'myInputData.testme'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR(''))) as getevts_mock:
      assertDiracFailsWith(self.moba.execute(),
                            'failed to get numberofevents from filecatalog', self)
      getevts_mock.assert_called_once_with('myInputData.testme')

  def test_execute_treatsteering_fails(self):
    self.moba.step_commons['SteeringFileVers'] = 'mySteerTestVers'
    self.moba.platform = 'TRestPlatformMine'
    with patch('%s.getSteeringFileDir' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_err_filedir_steer'))) as steerdir_mock:
      assertDiracFailsWith(self.moba.execute(),
                            'failed to locate steering files mysteertestvers', self)
      steerdir_mock.assert_called_once_with('TRestPlatformMine', 'mySteerTestVers')

  def test_execute_config_fails(self):
    self.moba.platform = 'PlatformMineTest'
    self.moba.workflow_commons['ClicConfigPackage'] = 'ClicConfigv102'
    self.moba.workflow_commons['SomeOtherEntry'] = 'Args'
    self.moba.workflow_commons['AndAnotherOne'] = 'Args'
    with patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_err'))) as cvmfs_mock, \
         patch('%s.getSoftwareFolder' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_err'))) as getsoft_mock:
      assertDiracFailsWith(self.moba.execute(),
                            'Failed to locate ClicConfigv102 as config dir', self)
      cvmfs_mock.assert_called_once_with('PlatformMineTest',
                                          ('clicconfig', 'v102'))
      getsoft_mock.assert_called_once_with('PlatformMineTest',
                                            'clicconfig', 'v102')

  def test_execute_runit_fails(self):
    with patch('%s.ModuleBase.runIt' % MODULE_NAME, new=Mock(return_value=S_ERROR('runit_fails_test_err'))):
      assertDiracFailsWith(self.moba.execute(), 'runit_fails_test_err', self)

  def test_execute_othercases(self):
    exists_dict = {'steering.file': False, './entry.steeringfile': False,
                    './failcopyingonthis': False, './testentry': True,
                    './lastfile.sf': False, './ildfile.entry': False,
                    './my.ild.file': True, './ild_failcopyhere': False,
                    './lastfile.ild': False}
    isdir_dict = {'steering/file/path/entry.steeringfile': True,
                   'steering/file/path/lastfile.sf': False,
                   'steering/file/path/failcopyingonthis': True,
                   'ild/test/configpath/ildfile.entry': True,
                   'ild/test/configpath/lastfile.ild': False,
                   'ild/test/configpath/ild_failcopyhere': True}
    listdir_dict = {'steering/file/path': [
        'entry.steeringfile', 'testentry', 'failcopyingonthis', 'lastfile.sf'],
        'ild/test/configpath': [
        'ildfile.entry', 'my.ild.file', 'ild_failcopyhere',
        'lastfile.ild'], '/test/cur/working/dir': []}
    self.moba.SteeringFile = '/dir/myfile/steering.file'
    self.moba.workflow_commons['ILDConfigPackage'] = 'ILDConfigv102'
    self.moba.step_commons['SteeringFileVers'] = 'mySteerTestVers'
    self.moba.platform = 'TRestPlatformMine'
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.getSteeringFileDir' % MODULE_NAME, new=Mock(return_value=S_OK('steering/file/path'))), \
         patch('%s.os.listdir' % MODULE_NAME, new=Mock(side_effect=lambda path: listdir_dict[path])) as listdir_mock, \
         patch('%s.os.path.isdir' % MODULE_NAME, new=Mock(side_effect=lambda path: isdir_dict[path])) as isdir_mock, \
         patch('%s.shutil.copytree' % MODULE_NAME, new=Mock(side_effect=[True, EnvironmentError('failed_copytree_testme'), True, EnvironmentError('failed_copytree_ildconfig')])) as copytree_mock, \
         patch('%s.shutil.copy2' % MODULE_NAME) as copy2_mock, \
         patch('%s.checkCVMFS' % MODULE_NAME, new=Mock(return_value=S_OK(['ild/test/configpath']))), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/test/cur/working/dir')) as getcwd_mock:
      result = self.moba.execute()
    assertDiracSucceedsWith_equals(result, None, self)
    assertMockCalls(copytree_mock, [('steering/file/path/entry.steeringfile', './entry.steeringfile'),
                                      ('steering/file/path/failcopyingonthis', './failcopyingonthis'),
                                      ('ild/test/configpath/ildfile.entry', './ildfile.entry'),
                                      ('ild/test/configpath/ild_failcopyhere', './ild_failcopyhere')], self)
    assertMockCalls(copy2_mock, [('steering/file/path/lastfile.sf', './lastfile.sf'),
                                   ('ild/test/configpath/lastfile.ild', './lastfile.ild')], self)
    assertMockCalls(listdir_mock, ['steering/file/path', 'ild/test/configpath', '/test/cur/working/dir'], self)
    assertMockCalls(isdir_mock, ['steering/file/path/entry.steeringfile',
                                   'steering/file/path/failcopyingonthis',
                                   'steering/file/path/lastfile.sf',
                                   'ild/test/configpath/ildfile.entry',
                                   'ild/test/configpath/ild_failcopyhere',
                                   'ild/test/configpath/lastfile.ild'], self)
    getcwd_mock.assert_called_once_with()

  def test_setappstat(self):
    self.moba.jobID = 24986
    log_mock = Mock()
    self.moba.log = log_mock
    report_mock = Mock()
    report_mock.setApplicationStatus.return_value = S_OK('mytest_success!!!')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracSucceedsWith_equals(self.moba.setApplicationStatus('my_test_status'),
                                    'mytest_success!!!', self)
    self.assertFalse(log_mock.called)
    self.assertFalse(log_mock.warn.called)
    self.assertFalse(log_mock.err.called)
    report_mock.setApplicationStatus.assert_called_once_with('my_test_status', True)

  def test_setappstat_local(self):
    self.moba.jobID = 0
    assertDiracSucceedsWith_equals(self.moba.setApplicationStatus('my_test_status'),
                                    'JobID not defined', self)

  def test_setappstat_noreporter(self):
    self.moba.jobID = 24986
    assertDiracSucceedsWith_equals(self.moba.setApplicationStatus('my_test_status'),
                                    'No reporting tool given', self)

  def test_setappstat_setting_fails(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.setApplicationStatus.return_value = S_ERROR('failed setting appstat_testme')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracFailsWith(self.moba.setApplicationStatus('my_test_status'),
                          'failed setting appstat_testme', self)
    report_mock.setApplicationStatus.assert_called_once_with('my_test_status', True)
    self.log_mock.warn.assert_called_once_with('failed setting appstat_testme')

  def test_sendstoredstatinfo(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.sendStoredStatusInfo.return_value = S_OK('mytest_success!!!')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracSucceedsWith_equals(self.moba.sendStoredStatusInfo(), 'mytest_success!!!', self)
    self.assertFalse(self.log_mock.called)
    self.assertFalse(self.log_mock.warn.called)
    self.assertFalse(self.log_mock.error.called)
    report_mock.sendStoredStatusInfo.assert_called_once_with()

  def test_sendstoredstatinfo_local(self):
    self.moba.jobID = 0
    assertDiracSucceedsWith_equals(self.moba.sendStoredStatusInfo(),
                                    'JobID not defined', self)

  def test_sendstoredstatinfo_noreporter(self):
    self.moba.jobID = 24986
    assertDiracSucceedsWith_equals(self.moba.sendStoredStatusInfo(),
                                    'No reporting tool given', self)

  def test_sendstoredstatinfo_setting_fails(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.sendStoredStatusInfo.return_value = S_ERROR('failed setting appstat_testme')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracFailsWith(self.moba.sendStoredStatusInfo(), 'failed setting appstat_testme', self)
    report_mock.sendStoredStatusInfo.assert_called_once_with()
    self.log_mock.error.assert_called_once_with('failed setting appstat_testme')

  def test_setjobparameter(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.setJobParameter.return_value = S_OK('mytest_success!!!')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracSucceedsWith_equals(self.moba.setJobParameter('mytestName', 135),
                                    'mytest_success!!!', self)
    self.assertFalse(self.log_mock.called)
    self.assertFalse(self.log_mock.warn.called)
    self.assertFalse(self.log_mock.error.called)
    report_mock.setJobParameter.assert_called_once_with('mytestName', '135', True)

  def test_setjobparameter_local(self):
    self.moba.jobID = 0
    assertDiracSucceedsWith_equals(self.moba.setJobParameter('mytestName', 193),
                                    'JobID not defined', self)

  def test_setjobparameter_noreporter(self):
    self.moba.jobID = 24986
    assertDiracSucceedsWith_equals(self.moba.setJobParameter('myTEstName', 9813),
                                    'No reporting tool given', self)

  def test_setjobparameter_setting_fails(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.setJobParameter.return_value = S_ERROR('failed setting appstat_testme')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracFailsWith(self.moba.setJobParameter('parameterTestName', 984),
                          'failed setting appstat_testme', self)
    report_mock.setJobParameter.assert_called_once_with('parameterTestName', '984', True)
    self.log_mock.warn.assert_called_once_with('failed setting appstat_testme')

  def test_sendstoredjobparameters(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.sendStoredJobParameters.return_value = S_OK('mytest_success!!!')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracSucceedsWith_equals(self.moba.sendStoredJobParameters(),
                                    'mytest_success!!!', self)
    self.assertFalse(self.log_mock.called)
    self.assertFalse(self.log_mock.warn.called)
    self.assertFalse(self.log_mock.error.called)
    report_mock.sendStoredJobParameters.assert_called_once_with()

  def test_sendstoredjobparameters_local(self):
    self.moba.jobID = 0
    assertDiracSucceedsWith_equals(self.moba.sendStoredJobParameters(),
                                    'JobID not defined', self)

  def test_sendstoredjobparameters_noreporter(self):
    self.moba.jobID = 24986
    assertDiracSucceedsWith_equals(self.moba.sendStoredJobParameters(),
                                    'No reporting tool given', self)

  def test_sendstoredjobparameters_setting_fails(self):
    self.moba.jobID = 24986
    report_mock = Mock()
    report_mock.sendStoredJobParameters.return_value = S_ERROR('failed setting appstat_testme')
    self.moba.workflow_commons['JobReport'] = report_mock
    assertDiracFailsWith(self.moba.sendStoredJobParameters(),
                          'failed setting appstat_testme', self)
    report_mock.sendStoredJobParameters.assert_called_once_with()
    self.log_mock.error.assert_called_once_with('failed setting appstat_testme')

  def test_setfilestatus_useexistingfilereport(self):
    report_mock = Mock()
    report_mock.setFileStatus.return_value = S_OK('my_report_test_returnval')
    self.moba.workflow_commons['FileReport'] = report_mock
    assertDiracSucceedsWith_equals(self.moba.setFileStatus('production', 'lfn', 'status'),
                                    'my_report_test_returnval', self)
    assertEqualsImproved(self.moba.workflow_commons['FileReport'], report_mock, self)

  def test_setfilestatus(self):
    report_mock = Mock()
    report_mock.setFileStatus.return_value = S_OK('other_my_report_test')
    with patch('%s.FileReport' % MODULE_NAME, new=Mock(return_value=report_mock)):
      assertDiracSucceedsWith_equals(self.moba.setFileStatus('production', 'lfn', 'status'),
                                      'other_my_report_test', self)
      assertEqualsImproved(self.moba.workflow_commons['FileReport'], report_mock, self)

  def test_setfilestatus_fails_useexistingfilereport(self):
    report_mock = Mock()
    report_mock.setFileStatus.return_value = S_ERROR('test_setfilestat_err')
    self.moba.workflow_commons['FileReport'] = report_mock
    assertDiracFailsWith(self.moba.setFileStatus('production', 'lfn', 'status'),
                          'test_setfilestat_err', self)
    assertEqualsImproved(self.moba.workflow_commons['FileReport'], report_mock, self)
    self.log_mock.warn.assert_called_once_with('test_setfilestat_err')

  def test_setfilestatus_fails(self):
    report_mock = Mock()
    report_mock.setFileStatus.return_value = S_ERROR('test_setfile_staterr')
    with patch('%s.FileReport' % MODULE_NAME, new=Mock(return_value=report_mock)):
      assertDiracFailsWith(self.moba.setFileStatus('production', 'lfn', 'status'),
                            'test_setfile_staterr', self)
      assertEqualsImproved(self.moba.workflow_commons['FileReport'], report_mock, self)
      self.log_mock.warn.assert_called_once_with('test_setfile_staterr')

  def test_getcandidatefiles(self):
    exists_dict = {'testfile_allworks.stdhep': True, 'testfile_notlocal.txt': False}
    self.moba.ignoreapperrors = True
    mytest_outputlist = [{
        'outputFile': 'testfile_allworks.stdhep', 'outputDataSE': 'testSE_dip4_allgood',
        'outputPath': '/test/clic/ilc/mytestfile.txt'}, {
        'outputFile': 'failhere', 'outputDataSE': ''}, {
        'outputFile': 'testfile_notlocal.txt', 'outputDataSE': 'no_se',
        'outputPath': '/test/clic/ilc/otherdir/newfile.txt'}]
    mylfns = ['testfile_allworks.stdhep', 'ignorethis', 'testfile_notlocal.txt']
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])):
      result = self.moba.getCandidateFiles(mytest_outputlist, mylfns, 'dummy_file_mask')
      assertDiracSucceedsWith_equals(result, {'testfile_allworks.stdhep': {
          'lfn': 'testfile_allworks.stdhep', 'path': '/test/clic/ilc/mytestfile.txt',
          'workflowSE': 'testSE_dip4_allgood'}}, self)

  def test_getcandidatefiles_filenametoolong(self):
    mytest_outputlist = [{'outputFile': 'eruighnegjmneroiljger89igujmnerjhvreikvnmer9fig8erjg89iuerjhguie5hgieu7hg893j4tf4iufnugfyrhbgyukbfruwjfhwiefjhuiewfjwenfiuewnfieuwhuifrweuijfiuwerjhuiwer',
                           'outputDataSE': 'testSE_dip4_allgood', 'outputPath': '/test/clic/ilc/mytestfile.txt'}, {'outputFile': 'failhere', 'outputDataSE': ''}]
    mylfns = [
        'eruighnegjmneroiljger89igujmnerjhvreikvnmer9fig8erjg89iuerjhguie5hgieu7hg893j4tf4iufnugfyrhbgyukbfruwjfhwiefjhuiewfjwenfiuewnfieuwhuifrweuijfiuwerjhuiwer',
        'ignorethis']
    assertDiracFailsWith(self.moba.getCandidateFiles(
        mytest_outputlist, mylfns, 'dummy_file_mask'), 'filename too long', self)

  def test_getcandidatefiles_lfntoolong(self):
    mytest_outputlist = [{
        'outputFile': 'testfile_dirstoolong.stdhep', 'outputDataSE': 'testSE_dip4_allgood',
        'outputPath': '/test/clic/ilc/mytestfile.txt'}, {
        'outputFile': 'failhere', 'outputDataSE': ''}]
    mylfns = ['esaiujf/oijkrgrmwg/oirwgjmoiwrg/oijefiouwef/dir/oiejfmwseroigfujwfguiwefmviwfweoifkmiwoe/oieujguimeosifkmespokfsoeifkjoisuejfsef/soiuejfuisejfosiekfoisejfiusejfoisekfjoisejuguisehngusefjoisefkfsefjmsi/eiujfmeiowfmefkjeoifjiuenfenfj/feiosjkfoiesfksepoflsefpolsefiokseiufnmjef/fueinsfsnejfhsnjhefsjhebfjshebfsenfseifnsoiefkjseoidejiuesjndeqniuwejqoiwjeiqwmdwkajndawnduaidjaiowdjiawd/duiwandqiuodjqiwodjqownuqnfrqujrnjqwrqweioqmwdoiqmid/testfile_dirstoolong.stdhep',
               'ignorethis']
    assertDiracFailsWith(self.moba.getCandidateFiles(
        mytest_outputlist, mylfns, 'dummy_file_mask'), 'lfn too long', self)

  def test_getcandidatefiles_missinglocally(self):
    exists_dict = {'dir/testfile_allworks.stdhep': True, 'testfile_notlocal.txt': False}
    self.moba.ignoreapperrors = False
    mytest_outputlist = [
        {'outputFile': 'dir/testfile_allworks.stdhep',
         'outputDataSE': 'testSE_dip4_allgood',
         'outputPath': '/test/clic/ilc/mytestfile.txt'}, {
            'outputFile': 'failhere', 'outputDataSE': ''}, {
            'outputFile': 'testfile_notlocal.txt', 'outputDataSE': 'no_se',
            'outputPath': '/test/clic/ilc/otherdir/newfile.txt'}]
    mylfns = ['testfile_allworks.stdhep', 'ignorethis', 'testfile_notlocal.txt']
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])):
      result = self.moba.getCandidateFiles(mytest_outputlist, mylfns, 'dummy_file_mask')
      assertDiracFailsWith(result, 'output data not found', self)

  def test_getfilemetadata(self):
    # Can in its current form never return S_ERROR, filedict is set when
    # adding new element to final and GUID is set in the metadata dict for
    # each element that might get added
    guid_dict = {'testfile_allworks.stdhep': 'test_myGuid_1', 'myothertest_file': 'test_myGuid_2'}
    size_dict = {'testfile_allworks.stdhep': 24852, 'myothertest_file': 948524}
    adler_dict = {'testfile_allworks.stdhep': '9803531', 'myothertest_file': 'checksum1230#'}
    with patch('%s.makeGuid' % MODULE_NAME, new=Mock(side_effect=lambda path: guid_dict[path])) as guid_mock, \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(side_effect=lambda path: size_dict[path])), \
         patch('%s.fileAdler' % MODULE_NAME, new=Mock(side_effect=lambda path: adler_dict[path])), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/cur/working/test/')):
      candidateFiles = {'testfile_allworks.stdhep': {
          'lfn': 'testfile_allworks.stdhep', 'path': '/test/clic/ilc/mytestfile.txt',
          'workflowSE': 'testSE_dip4_allgood'}, 'myothertest_file': {
          'lfn': 'myothertest_file', 'path': '/dir/clid/user/myothertestfile.txt',
          'workflowSE': 'CERN_dip4_testme'}}
      result = self.moba.getFileMetadata(candidateFiles)
      expected_dict = {
          'testfile_allworks.stdhep': {
              'filedict': {
                  'Status': 'Waiting', 'ADLER32': '9803531', 'ChecksumType': 'ADLER32',
                  'Checksum': '9803531', 'LFN': 'testfile_allworks.stdhep',
                  'GUID': 'test_myGuid_1', 'Addler': '9803531', 'Size': 24852
                  }, 'lfn': 'testfile_allworks.stdhep',
              'localpath': '/cur/working/test//testfile_allworks.stdhep',
              'workflowSE': 'testSE_dip4_allgood', 'path': '/test/clic/ilc/mytestfile.txt',
              'GUID': 'test_myGuid_1'}, 'myothertest_file': {
              'filedict': {
                  'Status': 'Waiting', 'ADLER32': 'checksum1230#',
                  'ChecksumType': 'ADLER32', 'Checksum': 'checksum1230#',
                  'LFN': 'myothertest_file', 'GUID': 'test_myGuid_2',
                  'Addler': 'checksum1230#', 'Size': 948524}, 'lfn': 'myothertest_file',
              'localpath': '/cur/working/test//myothertest_file',
              'workflowSE': 'CERN_dip4_testme',
              'path': '/dir/clid/user/myothertestfile.txt',
              'GUID': 'test_myGuid_2'}}
      assertDiracSucceedsWith_equals(result, expected_dict, self)
      assertMockCalls(guid_mock, ['testfile_allworks.stdhep', 'myothertest_file'], self)

  def test_resolveinputvars(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['InputData'] = 'myinputData1Test;myinputData2Test;myinputData3Test'
    mb.step_commons['InputFile'] = 'myInputTestFile1;TestinputFile2'
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))):
      result = mb.resolveInputVariables()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)

  def test_resolveinputvars_inputfile_variation(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['InputData'] = ['myinputData1Test', 'myinputData2Test',
                                        'LFN:myinputData3Test']
    mb.step_commons['InputFile'] = ''
    mb.OutputFile = 'myTestOutputFile'
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_OK({'AdditionalMeta': {}, 'nbevts': 9824}))):
      result = mb.resolveInputVariables()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)

  def test_resolveinputvarsinputfile_var2(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['InputData'] = ''
    # mb.workflow_commons['ParametricInputData'] = ''  test mit parametric+normal inputdata
    mb.step_commons['InputFile'] = []
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))):
      result = mb.resolveInputVariables()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)

  def test_resolveinputvars_parametricdata(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['InputData'] = 'myinputData1Test;myinputData2Test;myinputData3Test'
    mb.workflow_commons['ParametricInputData'] = 'paramData1Test;LFN:TestParamData2;3DataTest'
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))), \
         patch('%s.ModuleBase.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_failthisappspec'))):
      result = mb.resolveInputVariables()
      assertDiracFailsWith(result, 'test_failthisappspec', self)

  def test_resolveinputvars_parametricdata_1(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['ParametricInputData'] = ''
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))), \
         patch('%s.ModuleBase.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_failthisappspec'))):
      result = mb.resolveInputVariables()
      assertDiracFailsWith(result, 'test_failthisappspec', self)

  def test_resolveinputvars_parametricdata_2(self):
    mb = self.moba
    mb.workflow_commons['IS_PROD'] = True
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['ParametricInputData'] = ['myParamEntry1',
                                                   'LFN:someoTherEntry',
                                                   'LFN:dontforgetme']
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))), \
         patch('%s.ModuleBase.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_failthisappspec'))):
      result = mb.resolveInputVariables()
      assertDiracFailsWith(result, 'test_failthisappspec', self)

  def test_resolveinputvars_getnbevts_fails(self):
    mb = self.moba
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['InputData'] = 'myinputData1Test;myinputData2Test;myinputData3Test'
    mb.workflow_commons['ParametricInputData'] = ['myParamEntry1',
                                                   'LFN:someoTherEntry',
                                                   'LFN:dontforgetme']
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_ERROR('getNbEvtsFails_test'))), \
         patch('%s.ModuleBase.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_OK('bla'))):
      result = mb.resolveInputVariables()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)
      print(self.log_mock.mock_calls)
      self.log_mock.warn.assert_called_once_with(
          'Failed to get NumberOfEvents from FileCatalog, but this is not a production job')

  def test_resolveinputvars_getnbevts_zero(self):
    mb = self.moba
    mb.workflow_commons['PRODUCTION_ID'] = 13412
    mb.workflow_commons['SystemConfig'] = 'myTestPlatform'
    mb.workflow_commons['StartFrom'] = 94
    mb.workflow_commons['NbOfEvts'] = 458
    mb.workflow_commons['ParametricInputData'] = ['myParamEntry1',
                                                   'LFN:someoTherEntry',
                                                   'LFN:dontforgetme']
    mb.workflow_commons['InputData'] = 'myinputData1Test;myinputData2Test;myinputData3Test'
    mb.OutputFile = None
    mb.step_commons['OutputFile'] = 'myTestOFile.txt'
    with patch('%s.getNumberOfEvents' % MODULE_NAME, new=Mock(return_value=S_OK({'nbevts': 0, 'AdditionalMeta': {}}))), \
         patch('%s.ModuleBase.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_OK('bla'))):
      result = mb.resolveInputVariables()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)

  def test_finalstatusreport(self):
    self.moba.ignoreapperrors = False
    with patch('%s.ModuleBase.setApplicationStatus' % MODULE_NAME) as appstat_mock:
      assertDiracSucceedsWith(self.moba.finalStatusReport(0), ' Successful', self)
      appstat_mock.assert_called_once_with('  Successful')
      self.assertFalse(self.log_mock.error.called)

  def test_finalstatusreport_appfailed(self):
    self.moba.ignoreapperrors = False
    with patch('%s.ModuleBase.setApplicationStatus' % MODULE_NAME) as appstat_mock:
      assertDiracFailsWith(self.moba.finalStatusReport(1), '', self)
      appstat_mock.assert_called_once_with(' exited With Status 1')
      self.assertTrue(self.log_mock.error.called)

  def test_finalstatusreport_ignorefail(self):
    self.moba.ignoreapperrors = True
    with patch('%s.ModuleBase.setApplicationStatus' % MODULE_NAME) as appstat_mock:
      assertDiracSucceedsWith(self.moba.finalStatusReport(1), ' exited With Status 1', self)
      appstat_mock.assert_called_once_with(' exited With Status 1')
      self.assertTrue(self.log_mock.error.called)

  def test_generatefailover(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 1
    container_mock.toJSON.return_value = S_OK('myjsonfilecontent_testme')
    reqval_mock = Mock()
    reqval_mock.validate.return_value = S_OK('ok')
    reqval_mock.getJSONData.return_value = True
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock = Mock()
    report_mock.generateForwardDISET.return_value = S_OK(request_mock)
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    with patch('%s.open' % MODULE_NAME, mock_open()) as mo, \
         patch('%s.RequestValidator' % MODULE_NAME, new=Mock(return_value=reqval_mock)), \
         patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)):
      result = self.moba.generateFailoverFile()
      assertDiracSucceeds(result, self)
      mo.assert_any_call('0_0_request.json', 'w')
      mo = mo()
      mo.write.assert_called_once_with('myjsonfilecontent_testme')

  def test_generatefailover_optionalfails(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 1
    container_mock.toJSON.return_value = S_OK('myjsonfilecontent_testme')
    container_mock.getDigest.return_value = S_ERROR('bla')
    reqval_mock = Mock()
    reqval_mock.validate.return_value = S_OK('ok')
    reqval_mock.getJSONData.return_value = True
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock = Mock()
    report_mock.generateForwardDISET.return_value = S_ERROR('test_faildiset')
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    accreport_mock = Mock()
    accreport_mock.commit.return_value = S_OK('')
    self.moba.workflowStatus['OK'] = False
    self.moba.workflow_commons['AccountingReport'] = accreport_mock
    self.moba.workflow_commons['ProductionOutputData'] = 'mylfn1;testmelfn2;'
    with patch('%s.open' % MODULE_NAME, mock_open()) as mo, \
         patch('%s.RequestValidator' % MODULE_NAME, new=Mock(return_value=reqval_mock)), \
         patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)), \
         patch('%s.ModuleBase.setApplicationStatus' % MODULE_NAME) as appstat_mock, \
         patch('%s.ModuleBase._cleanUp' % MODULE_NAME) as cleanup_mock:
      result = self.moba.generateFailoverFile()
      assertDiracSucceeds(result, self)
      mo.assert_any_call('0_0_request.json', 'w')
      mo = mo()
      mo.write.assert_called_once_with('myjsonfilecontent_testme')
      appstat_mock.assert_called_once_with('Creating Removal Requests')
      cleanup_mock.assert_called_once_with(['mylfn1', 'testmelfn2', ''])

  def test_generatefailover_commit_fails(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 1
    container_mock.toJSON.return_value = S_OK('myjsonfilecontent_testme')
    container_mock.getDigest = S_ERROR('bla')
    reqval_mock = Mock()
    reqval_mock.validate.return_value = S_OK('ok')
    reqval_mock.getJSONData.return_value = True
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock = Mock()
    report_mock.generateForwardDISET.return_value = S_ERROR('test_faildiset')
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    accreport_mock = Mock()
    accreport_mock.commit.return_value = S_ERROR('test_commiterr')
    self.moba.workflowStatus['OK'] = False
    self.moba.workflow_commons['AccountingReport'] = accreport_mock
    self.moba.workflow_commons['ProductionOutputData'] = 'mylfn1;testmelfn2;'
    with patch('%s.RequestValidator' % MODULE_NAME, new=Mock(return_value=reqval_mock)), \
         patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)):
      result = self.moba.generateFailoverFile()
      assertDiracFailsWith(result, 'test_commiterr', self)

  def test_generatefailover_norequests(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 0
    container_mock.toJSON.return_value = S_OK('myjsonfilecontent_testme')
    report_mock = Mock()
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock.generateForwardDISET.return_value = S_OK(request_mock)
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    with patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)):
      result = self.moba.generateFailoverFile()
      assertDiracSucceeds(result, self)
      self.log_mock.info.assert_any_call('No Requests to process ')

  def test_generatefailover_validatefails(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 1
    container_mock.toJSON.return_value = S_OK('myjsonfilecontent_testme')
    reqval_mock = Mock()
    reqval_mock.validate.return_value = S_ERROR('test_fail-validate')
    reqval_mock.getJSONData.return_value = True
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock = Mock()
    report_mock.generateForwardDISET.return_value = S_OK(request_mock)
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    with patch('%s.RequestValidator' % MODULE_NAME, new=Mock(return_value=reqval_mock)), \
         patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)):
      with self.assertRaises(RuntimeError) as re:
        self.moba.generateFailoverFile()
      assertEqualsImproved(re.exception.args[0],
                           'Failover request is not valid: test_fail-validate',
                           self)

  def test_generatefailover_tojsonfails(self):
    container_mock = Mock()
    container_mock.__len__.return_value = 1
    container_mock.toJSON.return_value = S_ERROR('test_tojsonerror')
    reqval_mock = Mock()
    reqval_mock.validate.return_value = S_OK('ok')
    reqval_mock.getJSONData.return_value = True
    request_mock = Mock()
    request_mock.getJSONData.return_value = True
    report_mock = Mock()
    report_mock.generateForwardDISET.return_value = S_OK(request_mock)
    report_mock.getJSONData.return_value = True
    self.moba.jobReport = report_mock
    with patch('%s.RequestValidator' % MODULE_NAME, new=Mock(return_value=reqval_mock)), \
         patch('%s.ModuleBase._getRequestContainer' % MODULE_NAME, new=Mock(return_value=container_mock)):
      with self.assertRaises(RuntimeError) as re:
        self.moba.generateFailoverFile()
      assertEqualsImproved(re.exception.args[0], 'test_tojsonerror', self)

# TODO Fix this/remove tests?
  def test_redirectlogoutput(self):
    self.moba.eventstring = None
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(1, 'mytestmessage'))
      # if print_mock.getvalue() not in [ 'mytestmessage\n', '' ]:
      if print_mock.getvalue() not in ['mytestmessage\n']:
        self.fail('Suitable output not found')
      self.assertFalse(open_mock.called)
      assertEqualsImproved(self.moba.stdError, 'mytestmessage', self)

  def test_redirectlogoutput_unicode(self):
    self.moba.eventstring = None
    self.moba.applicationLog = 'mylogg.txt'
    with patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.moba.redirectLogOutput(1, '100 +- 10')  # plus minus
      self.moba.redirectLogOutput(1, u'100 ± 10')  # plus minus sign
      open_mock().write.assert_any_call("100 +- 10\n")
      open_mock().write.assert_any_call(u"100 ± 10\n")

  def test_redirectlogoutput_emptymsg(self):
    self.assertIsNone(self.moba.redirectLogOutput('fd', ''))

  def test_redirectlogoutput_default(self):
    self.moba.eventstring = 'testevent123'
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(
          0, 'testevent123 has happened! quick, print it!'))
      # if print_mock.getvalue() not in [ 'testevent123 has happened! quick, print it!\n', '' ]:
      if print_mock.getvalue() not in ['testevent123 has happened! quick, print it!\n']:
        assert False, print_mock.getvalue()
      self.assertFalse(open_mock.called)
      assertEqualsImproved(self.moba.stdError, '', self)

  def test_redirectlogoutput_writetofile(self):
    self.moba.eventstring = None
    self.moba.applicationLog = 'appLog.txt'
    self.moba.excludeAllButEventString = False
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(1, 'mytestmessage'))
      if print_mock.getvalue() not in ['mytestmessage\n', '']:
        self.fail('Suitable output not found')
      open_mock.assert_any_call('appLog.txt', 'at')
      open_mock = open_mock()
      open_mock.write.assert_called_once_with('mytestmessage\n')
      assertEqualsImproved(self.moba.stdError, 'mytestmessage', self)

  def test_redirectlogoutput_writetofile_2(self):
    self.moba.eventstring = ''
    self.moba.applicationLog = 'appLog.txt'
    self.moba.excludeAllButEventString = True
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(0, 'mytestmessage'))
      assert print_mock.getvalue() == ''
      open_mock.assert_any_call('appLog.txt', 'at')
      self.assertFalse(open_mock().called)

  def test_redirectlogoutput_writetofile_3(self):
    self.moba.eventstring = ['somepattern', 'otherpattern']
    self.moba.applicationLog = 'appLog.txt'
    self.moba.excludeAllButEventString = True
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(0, 'mytestmessage'))
      assert print_mock.getvalue() == ''
      open_mock.assert_any_call('appLog.txt', 'at')
      open_mock = open_mock()
      self.assertFalse(open_mock.write.called)

  def test_redirectlogoutput_writetofile_4(self):
    self.moba.eventstring = ['ignorethis', 'deletethis', 'specialTestEvent']
    self.moba.applicationLog = 'appLog.txt'
    self.moba.excludeAllButEventString = True
    with patch('sys.stdout', new_callable=StringIO) as print_mock, \
         patch('%s.open' % MODULE_NAME, mock_open()) as open_mock:
      self.assertIsNone(self.moba.redirectLogOutput(1, '1390specialTestEvente89f'))
      if print_mock.getvalue() not in ['1390specialTestEvente89f\n', '']:
        self.fail('Suitable output not found')
      open_mock.assert_any_call('appLog.txt', 'at')
      open_mock = open_mock()
      open_mock.write.assert_called_once_with('1390specialTestEvente89f\n')
      assertEqualsImproved(self.moba.stdError, '1390specialTestEvente89f', self)

  def test_cleanup(self):
    file_mock = Mock()
    with patch('%s.File' % MODULE_NAME, new=Mock(return_value=file_mock)):
      result = self.moba._cleanUp(['lfnlist'])  # pylint: disable=protected-access
      assertDiracSucceeds(result, self)
      self.assertIsNotNone(self.moba.workflow_commons['Request'])

  def test_logworkingdirectory(self):
    with patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=S_OK([0, 'balblabal', '']))):
      self.moba.logWorkingDirectory()
      self.log_mock.info.assert_any_call('balblabal')
      self.assertFalse(self.log_mock.error.called)

  def test_logworkingdirectory_fails_1(self):
    with patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=S_OK([1, 'balblabal', 'test_err']))):
      self.moba.logWorkingDirectory()
      self.assertTrue(self.log_mock.error.called)
      self.assertFalse(self.log_mock.info.called)

  def test_logworkingdirectory_fails_2(self):
    error = S_ERROR([0, 'balblabal', 'myerrormsg'])
    error['Value'] = 'efopikif'
    with patch('%s.shellCall' % MODULE_NAME, new=Mock(return_value=error)):
      self.moba.logWorkingDirectory()
      self.assertTrue(self.log_mock.error.called)
      self.assertFalse(self.log_mock.info.called)

# TODO Check for appropriately set values



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""List of files in the SteeringFiles tar ball.

:author: S. Poss
:since: Nov 8, 2011
"""

from __future__ import absolute_import
import os

from DIRAC import S_OK, S_ERROR, gLogger

LOG = gLogger.getSubLogger(__name__)


def Exists(myfile, platform=None, configversion=None):
  """check if the file exists in the tarball First based on a list of files.

  Also checks if CVMFS is used for the configuration package (ILDConfig) set by
  :func:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setILDConfig`
  If CVMFS is not available in the submission machine, but ILDConfig is on CVMFS we assume the file exists.

  :param str myfile: filename to be checked
  :param str platform: requested platform, optional
  :param str configversion: ILDConfig version defined for the :class:`~ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob`
  :returns: S_OK/S_ERROR
  """
  files = ["defaultClicCrossingAngle.mac", "clic_ild_cdr500.steer",
           "clic_ild_cdr.steer", "clic_cdr_prePandora.lcsim",
           "clic_cdr_postPandora.lcsim", "clic_cdr_prePandoraOverlay.lcsim",
           "clic_cdr_prePandoraOverlay_1400.0.lcsim",
           "clic_cdr_prePandoraOverlay_3000.0.lcsim",
           "clic_cdr_postPandoraOverlay.lcsim", "clic_ild_cdr.gear",
           "clic_ild_cdr500.gear", "clic_ild_cdr_steering_overlay.xml",
           "clic_ild_cdr_steering_overlay_3000.0.xml",
           "clic_ild_cdr_steering_overlay_1400.0.xml",
           "clic_ild_cdr500_steering_overlay.xml",
           "clic_ild_cdr500_steering_overlay_350.0.xml",
           "clic_ild_cdr_steering.xml",
           "clic_ild_cdr500_steering.xml", "GearOutput.xml",
           'cuts_e1e1ff_500gev.txt',
           "cuts_e2e2ff_500gev.txt", 'cuts_qq_nunu_1400.txt',
           'cuts_e3e3nn_1400.txt',
           "cuts_e3e3_1400.txt", "cuts_e1e1e3e3_o_1400.txt",
           "cuts_aa_e3e3_o_1400.txt",
           "cuts_aa_e3e3nn_1400.txt", "cuts_aa_e2e2e3e3_o_1400.txt",
           "cuts_aa_e1e1e3e3_o_1400.txt",
           "defaultStrategies_clic_sid_cdr.xml",
           "defaultIlcCrossingAngle.mac",
           "defaultIlcCrossingAngleZSmearing320.mac",
           "defaultIlcCrossingAngleZSmearing225.mac",
           "sid_dbd_pandoraSettings.xml",
           "sid_dbd_postPandora.xml",
           "sid_dbd_prePandora.xml",
           "sid_dbd_prePandora_noOverlay.xml",
           "sid_dbd_vertexing.xml",
           "sidloi3.gear",
           "sidloi3_trackingStrategies.xml",
           "sidloi3_trackingStrategies_default.xml",
           "ild_00.gear",
           "ild_00_steering.xml",
           "ild_00.steer",
           "cuts_quarks_1400.txt", "cuts_taus_1400.txt",
           "cuts_h_gammaZ_1400.txt", "cuts_h_gammagamma_1400.txt",
           "cuts_h_mumu_3000.txt", ]
  if myfile in files:
    return S_OK()
  elif configversion is None or platform is None:
    return S_ERROR("File %s is not available locally nor in the software installation." % myfile)
  else:
    return _checkInCVMFS(myfile, platform, configversion)


def _checkInCVMFS(cFile, platform, configversion):
  """check if the file is available on cvmfs or not."""

  from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
  version = configversion.split("Config")[1]
  app = configversion.split("Config")[0] + "Config"
  configPath = Operations().getValue("/AvailableTarBalls/%(platform)s/%(app)s/%(version)s/CVMFSPath" %
                                         dict(version=version,
                                               platform=platform,
                                               app=app.lower(),
                                             ),
                                      "")
  if not configPath:
    return S_ERROR("Cannot get CVMFSPath for this ILDConfig version: %s " % version)

  # check if cvmfs exists on this machine, if not we guess the person knows what they are doing
  if not os.path.exists("/cvmfs"):
    LOG.warn("CMVFS does not exist on this machine, cannot check for file existance.")
    return S_OK()

  if os.path.exists(os.path.join(configPath, cFile)):
    LOG.info("Found file on CVMFS %s/%s" % (configPath, cFile))
    return S_OK()
  else:
    LOG.error("Cannot find file %s in cvmfs folder: %s  " % (cFile, configPath))
    return S_ERROR("Cannot find file on cvmfs")



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Module with functions to get proxies for specific groups.

Special function to get production Checks and potentially provides production proxy, called from :mod:`~ILCDIRAC.Interfaces.API.NewInterface.ProductionJob`

:since: Feb 10, 2011
:author: sposs, A.Sailer
"""
from __future__ import print_function
from __future__ import absolute_import
from subprocess import call

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

LOG = gLogger.getSubLogger(__name__)

__RCSID__ = "$Id$"


def getNewProxy(group="ilc_prod"):
  """Get a new production proxy.

  :param str group: dirac group of the proxy to get
  :returns: statuscode of the dirac-proxy-init call: 0 success, otherwise error!
  """
  LOG.notice('Getting production proxy ...')
  return call(['dirac-proxy-init', '-g', group])


def checkAndGetProdProxy():
  """Check if current proxy is a production one, and if not call the :any:`getNewProxy` method.

  .. deprecated:: v25r0p0 use checkOrGetGroupProxy

  :returns: :func:`S_OK() <DIRAC:DIRAC.Core.Utilities.ReturnValues.S_OK>`, :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_ERROR`
  """
  return checkOrGetGroupProxy(group="ilc_prod")


def checkOrGetGroupProxy(group):
  """Check if current proxy corresponds to the given group, and if not call the
  :any:`getNewProxy` method to obtain a proxy for this group

  :param group: dirac group of the desired proxy
  :type group: list or str
  :returns: :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_OK`,
     :func:`~DIRAC:DIRAC.Core.Utilities.ReturnValues.S_ERROR`
  """
  result = getProxyInfo()
  groups = group
  if not isinstance(groups, (list, set, tuple)):
    groups = [groups]
  if result['OK'] and 'group' in result['Value']:
    if result['Value']['group'] in groups:
      return S_OK(result['Value']['group'])
    else:
      LOG.error("You don't have an %s proxy, trying to get one..." % group)
  else:
    LOG.error("Error to get proxy information, trying to get proxy")

  if not len(groups) == 1:
    return S_ERROR("More than one proxy group possible, cannot continue, please get proper proxy")

  newProxyRetVal = getNewProxy(group=groups[0])
  if newProxyRetVal:  # != 0
    return S_ERROR("dirac-proxy-init failed")

  result = getProxyInfo()
  if result['OK'] and 'group' in result['Value']:
    if result['Value']['group'] in groups:
      return S_OK(result['Value']['group'])
    else:
      LOG.error('You do not have a valid group')
      return S_ERROR("Could not obtain valid group")
  elif result['OK'] and 'group' not in result['Value']:
    return S_ERROR("Could not obtain group information from proxy")

  LOG.error('Could not obtain proxy information: %s' % result['Message'])
  return result
